from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Union
from typing import Any


def identity(arg: Any = ..., *args, **kwargs) -> Any:
    ...


def inject_method(self, func: Callable[..., Any], name: str = ...) -> None:
    ...


def compatible(config: Dict[str, Any],
               func: Callable,
               start: int = ...,
               keywords: Union[bool, Iterable[str]] = ...) -> Dict[str, Any]:
    ...



from typing import List
from typing import Union
from os import PathLike
from typing import Dict
from _typeshed import Incomplete

__pitch__: str
POSIX: Incomplete
WIN32: Incomplete


def cmd(command: Union[str, List[str]],
        shell: bool = ...,
        detach: bool = ...,
        verbose: int = ...,
        tee: Union[bool, None] = ...,
        cwd: Union[str, PathLike, None] = ...,
        env: Union[Dict[str, str], None] = ...,
        tee_backend: str = ...,
        check: bool = ...,
        system: bool = ...,
        timeout: float = ...) -> dict:
    ...



from typing import Union
from os import PathLike
from typing import Iterable
from typing import List
from collections.abc import Generator

WIN32: bool
LINUX: bool
FREEBSD: bool
DARWIN: bool
POSIX: bool


def platform_data_dir() -> str:
    ...


def platform_config_dir() -> str:
    ...


def platform_cache_dir() -> str:
    ...


def get_app_data_dir(appname: str, *args) -> str:
    ...


def ensure_app_data_dir(appname: str, *args) -> str:
    ...


def get_app_config_dir(appname: str, *args) -> str:
    ...


def ensure_app_config_dir(appname: str, *args) -> str:
    ...


def get_app_cache_dir(appname: str, *args) -> str:
    ...


def ensure_app_cache_dir(appname: str, *args) -> str:
    ...


def find_exe(
    name: Union[str, PathLike],
    multi: bool = ...,
    path: Union[str, PathLike, Iterable[Union[str, PathLike]], None] = ...
) -> str | List[str] | None:
    ...


def find_path(name: Union[str, PathLike],
              path: Union[str, Iterable[Union[str, PathLike]]] = ...,
              exact: bool = ...) -> Generator[str, None, None]:
    ...



from typing import Tuple
from typing import Union
from ubelt.util_const import NoParamType
from typing import List
from typing import Optional
from typing import TypeVar

T = TypeVar("T")


def argval(key: Union[str, Tuple[str, ...]],
           default: Union[T, NoParamType] = ...,
           argv: Optional[List[str]] = ...) -> str | T:
    ...


def argflag(key: Union[str, Tuple[str, ...]], argv: List[str] = ...) -> bool:
    ...



"""
Currently this module provides one utility
:func:`ubelt.util_deprecate.schedule_deprecation` which allows a developer to
easily mark features in their libraries as deprecated.
"""


def schedule_deprecation(modname, name='?', type='?', migration='',
                         deprecate=None, error=None, remove=None,
                         warncls=DeprecationWarning, stacklevel=1):
    """
    Deprecation machinery to help provide users with a smoother transition.

    This function provides a concise way to mark a feature as deprecated by
    providing a description of the deprecated feature, documentation on how to
    migrate away from the deprecated feature, and the versions that the feature
    is scheduled for deprecation and eventual removal. Based on the version of
    the library and the specified schedule this function will either do
    nothing, emit a warning, or raise an error with helpful messages for both
    users and developers.

    Args:
        modname (str):
            The name of the underlying module associated with the feature to be
            deprecated. The module must already be imported and have a passable
            ``__version__`` attribute.

        name (str):
            The name of the feature to deprecate. This is usually a function or
            argument name.

        type (str):
            A description of what the feature is. This is not a formal type,
            but rather a prose description: e.g. "argument to my_func".

        migration (str):
            A description that lets users know what they should do instead of
            using the deprecated feature.

        deprecate (str | None):
            The version when the feature is officially deprecated and this
            function should start to emit a deprecation warning.
            Can also be the strings: "soon" or "now" if the timeline
            isnt perfectly defined.

        error (str | None):
            The version when the feature is officially no longer supported, and
            will start to raise a RuntimeError.
            Can also be the strings: "soon" or "now".

        remove (str | None):
            The version when the feature is completely removed. An
            AssertionError will be raised if this function is still present
            reminding the developer to remove the feature (or extend the remove
            version).  Can also be the strings: "soon" or "now".

        warncls (type):
            This is the category of warning to use. Defaults to
            :class:`DeprecationWarning`.

        stacklevel (int):
            The stacklevel can be used by wrapper functions to indicate where
            the warning is occurring.

    Returns:
        str : the constructed message

    Note:
        If deprecate, remove, or error is specified as "now" or a truthy value
        it will force that check to trigger immediately. If the value is
        "soon", then the check will not trigger.

    Note:
        The :class:`DeprecationWarning` is not visible by default.
        https://docs.python.org/3/library/warnings.html

    Example:
        >>> import ubelt as ub
        >>> import sys
        >>> import types
        >>> import pytest
        >>> dummy_module = sys.modules['dummy_module'] = types.ModuleType('dummy_module')
        >>> # When less than the deprecated version this does nothing
        >>> dummy_module.__version__ = '1.0.0'
        >>> ub.schedule_deprecation(
        ...     'dummy_module', 'myfunc', 'function', 'do something else',
        ...     deprecate='1.1.0', error='1.2.0', remove='1.3.0')
        >>> # But when the module version increases above the threshold,
        >>> # the warning is raised.
        >>> dummy_module.__version__ = '1.1.0'
        >>> with pytest.warns(DeprecationWarning):
        ...     msg = ub.schedule_deprecation(
        ...         'dummy_module', 'myfunc', 'function', 'do something else',
        ...         deprecate='1.1.0', error='1.2.0', remove='1.3.0')
        >>> print(msg)
        The "myfunc" function was deprecated in 1.1.0, will cause an error in
        1.2.0 and will be removed in 1.3.0. The current version is 1.1.0. do
        something else

    Example:
        >>> # Demo the various cases
        >>> import ubelt as ub
        >>> import sys
        >>> import types
        >>> import pytest
        >>> dummy_module = sys.modules['dummy_module'] = types.ModuleType('dummy_module')
        >>> # When less than the deprecated version this does nothing
        >>> dummy_module.__version__ = '1.1.0'
        >>> # Now this raises warning
        >>> with pytest.warns(DeprecationWarning):
        ...     dummy_module.__version__ = '1.1.0'
        ...     ub.schedule_deprecation(
        ...         'dummy_module', 'myfunc', 'function', 'do something else',
        ...         deprecate='1.1.0', error='1.2.0', remove='1.3.0')
        >>> # Now this raises an error for the user
        >>> with pytest.raises(RuntimeError):
        ...     dummy_module.__version__ = '1.2.0'
        ...     ub.schedule_deprecation(
        ...         'dummy_module', 'myfunc', 'function', 'do something else',
        ...         deprecate='1.1.0', error='1.2.0', remove='1.3.0')
        >>> # Now this raises an error for the developer
        >>> with pytest.raises(AssertionError):
        ...     dummy_module.__version__ = '1.3.0'
        ...     ub.schedule_deprecation(
        ...         'dummy_module', 'myfunc', 'function', 'do something else',
        ...         deprecate='1.1.0', error='1.2.0', remove='1.3.0')
        >>> # When no versions are specified, it simply emits the warning
        >>> with pytest.warns(DeprecationWarning):
        ...     dummy_module.__version__ = '1.1.0'
        ...     ub.schedule_deprecation(
        ...         'dummy_module', 'myfunc', 'function', 'do something else')
        >>> # Test with soon / now
        >>> with pytest.warns(Warning):
        ...     ub.schedule_deprecation(
        ...         'dummy_module', 'myfunc', 'function', 'do something else',
        ...         deprecate='now', error='soon', remove='soon', warncls=Warning)
        >>> # Test with truthy values
        >>> with pytest.raises(RuntimeError):
        ...     ub.schedule_deprecation(
        ...         'dummy_module', 'myfunc', 'function', 'do something else',
        ...         deprecate=True, error=1, remove=False)
        >>> # Test with No module
        >>> with pytest.warns(Warning):
        ...     ub.schedule_deprecation(
        ...         None, 'myfunc', 'function', 'do something else',
        ...         deprecate='now', error='soon', remove='soon', warncls=Warning)
        >>> # Test with No module
        >>> with pytest.warns(Warning):
        ...     ub.schedule_deprecation(
        ...         None, 'myfunc', 'function', 'do something else',
        ...         deprecate='now', error='2.0.0', remove='soon', warncls=Warning)
    """
    import sys
    import warnings
    from packaging.version import parse as Version

    if modname is not None:
        module = sys.modules[modname]
        current = Version(module.__version__)
    else:
        current = 'unknown'

    def _handle_when(when, default):
        if when is None:
            is_now = default
            when_str = ''
        elif isinstance(when, str):
            if when in {'soon', 'now'}:
                when_str = ' {}'.format(when)
                is_now = (when == 'now')
            else:
                when = Version(when)
                when_str = ' in {}'.format(when)
                if current == 'unknown':
                    is_now = default
                else:
                    is_now = current >= when
        else:
            is_now = bool(when)
            when_str = ''
        return is_now, when_str

    deprecate_now, deprecate_str = _handle_when(deprecate, default=True)
    remove_now, remove_str = _handle_when(remove, default=False)
    error_now, error_str = _handle_when(error, default=False)

    msg = (
        'The "{name}" {type} was deprecated{deprecate_str}, will cause '
        'an error{error_str} and will be removed{remove_str}. The current '
        'version is {current}. {migration}'
    ).format(**locals()).strip()
    if remove_now:
        raise AssertionError(
            'Forgot to remove deprecated: ' + msg + ' ' +
            'Remove the function, or extend the scheduled remove version.'
        )
    if error_now:
        raise RuntimeError(msg)

    if deprecate_now:
        warnings.warn(msg, warncls, stacklevel=1 + stacklevel)

    return msg



from typing import Union
from os import PathLike


def symlink(real_path: Union[str, PathLike],
            link_path: Union[str, PathLike],
            overwrite: bool = ...,
            verbose: int = ...) -> str | PathLike:
    ...



"""
Path and filesystem utilities.

The :class:`Path` object is an extension of :class:`pathlib.Path` that contains
extra convenience methods corresponding to the extra functional methods in this
module. (New in 0.11.0). See the class documentation for more details.

This module also defines functional path-related utilities, but moving forward
users should prefer using :class:`Path` over standalone functional methods. The
functions methods will still be available for the forseable future, but their
functionality is made redundant by :class:`Path`. For completeness these
functions are listed

The :func:`expandpath` function expands the tilde to $HOME and environment
variables to their values.

The :func:`augpath` function creates variants of an existing path without
having to spend multiple lines of code splitting it up and stitching it back
together.

The :func:`shrinkuser` function replaces your home directory with a tilde.

The :func:`userhome` function reports the home directory of the current user of
the operating system.

The :func:`ensuredir` function operates like ``mkdir -p`` in unix.

Note:
    In the future the part of this module that defines Path may be renamed to
    util_pathlib.
"""
from os.path import (
    dirname, exists, expanduser, expandvars, join, normpath, split, splitext,
)
import os
import sys
from ubelt import util_io
import pathlib
import warnings


__all__ = [
    'Path', 'TempDir', 'augpath', 'shrinkuser', 'userhome', 'ensuredir',
    'expandpath',
]


def augpath(path, suffix='', prefix='', ext=None, tail='', base=None,
            dpath=None, relative=None, multidot=False):
    """
    Create a new path with a different extension, basename, directory, prefix,
    and/or suffix.

    A prefix is inserted before the basename. A suffix is inserted
    between the basename and the extension. The basename and extension can be
    replaced with a new one. Essentially a path is broken down into components
    (dpath, base, ext), and then recombined as (dpath, prefix, base, suffix,
    ext) after replacing any specified component.

    Args:
        path (str | PathLike): a path to augment

        suffix (str):
            placed between the basename and extension
            Note: this is referred to as stemsuffix in :func:`ub.Path.augment`.

        prefix (str):
            placed in front of the basename

        ext (str | None):
            if specified, replaces the extension

        tail (str | None):
            If specified, appends this text to the extension

        base (str | None):
            if specified, replaces the basename without extension.
            Note: this is referred to as stem in :func:`ub.Path.augment`.

        dpath (str | PathLike | None):
            if specified, replaces the specified "relative" directory, which by
            default is the parent directory.

        relative (str | PathLike | None):
            Replaces ``relative`` with ``dpath`` in ``path``.
            Has no effect if ``dpath`` is not specified.
            Defaults to the dirname of the input ``path``.
            *experimental* not currently implemented.

        multidot (bool): Allows extensions to contain multiple
            dots. Specifically, if False, everything after the last dot in the
            basename is the extension. If True, everything after the first dot
            in the basename is the extension.

    Returns:
        str: augmented path

    Example:
        >>> import ubelt as ub
        >>> path = 'foo.bar'
        >>> suffix = '_suff'
        >>> prefix = 'pref_'
        >>> ext = '.baz'
        >>> newpath = ub.augpath(path, suffix, prefix, ext=ext, base='bar')
        >>> print('newpath = %s' % (newpath,))
        newpath = pref_bar_suff.baz

    Example:
        >>> from ubelt.util_path import *  # NOQA
        >>> augpath('foo.bar')
        'foo.bar'
        >>> augpath('foo.bar', ext='.BAZ')
        'foo.BAZ'
        >>> augpath('foo.bar', suffix='_')
        'foo_.bar'
        >>> augpath('foo.bar', prefix='_')
        '_foo.bar'
        >>> augpath('foo.bar', base='baz')
        'baz.bar'
        >>> augpath('foo.tar.gz', ext='.zip', multidot=True)
        foo.zip
        >>> augpath('foo.tar.gz', ext='.zip', multidot=False)
        foo.tar.zip
        >>> augpath('foo.tar.gz', suffix='_new', multidot=True)
        foo_new.tar.gz
        >>> augpath('foo.tar.gz', suffix='_new', tail='.cache', multidot=True)
        foo_new.tar.gz.cache
    """
    stem = base  # new nomenclature

    # Breakup path
    if relative is None:
        orig_dpath, fname = split(path)
    else:  # nocover
        # if path.startswith(relative):
        #     orig_dpath = relative
        #     fname = relpath(path, relative)
        # else:
        #     orig_dpath, fname = split(path)
        raise NotImplementedError('Not implemented yet')

    if multidot:
        # The first dot defines the extension
        parts = fname.split('.', 1)
        orig_base = parts[0]
        orig_ext = '' if len(parts) == 1 else '.' + parts[1]
    else:
        # The last dot defines the extension
        orig_base, orig_ext = splitext(fname)
    # Replace parts with specified augmentations
    if dpath is None:
        dpath = orig_dpath
    if ext is None:
        ext = orig_ext
    if stem is None:
        stem = orig_base
    # Recombine into new path
    new_fname = ''.join((prefix, stem, suffix, ext, tail))
    newpath = join(dpath, new_fname)
    return newpath


def userhome(username=None):
    """
    Returns the path to some user's home directory.

    Args:
        username (str | None):
            name of a user on the system. If not specified, the current user is
            inferred.

    Returns:
        str: userhome_dpath - path to the specified home directory

    Raises:
        KeyError: if the specified user does not exist on the system

        OSError: if username is unspecified and the current user cannot be
            inferred

    Example:
        >>> from ubelt.util_path import *  # NOQA
        >>> import ubelt as ub
        >>> import getpass
        >>> username = getpass.getuser()
        >>> userhome_target = expanduser('~')
        >>> userhome_got1 = ub.userhome()
        >>> userhome_got2 = ub.userhome(username)
        >>> print(f'username={username}')
        >>> print(f'userhome_got1={userhome_got1}')
        >>> print(f'userhome_got2={userhome_got2}')
        >>> print(f'userhome_target={userhome_target}')
        >>> assert userhome_got1 == userhome_target
        >>> assert userhome_got2 == userhome_target
    """
    if username is None:
        # get home directory for the current user
        if 'HOME' in os.environ:
            userhome_dpath = os.environ['HOME']
        else:  # nocover
            if sys.platform.startswith('win32'):
                # win32 fallback when HOME is not defined
                if 'USERPROFILE' in os.environ:
                    userhome_dpath = os.environ['USERPROFILE']
                elif 'HOMEPATH' in os.environ:
                    drive = os.environ.get('HOMEDRIVE', '')
                    userhome_dpath = join(drive, os.environ['HOMEPATH'])
                else:
                    raise OSError("Cannot determine the user's home directory")
            else:
                # posix fallback when HOME is not defined
                import pwd
                userhome_dpath = pwd.getpwuid(os.getuid()).pw_dir
    else:
        # A specific user directory was requested
        if sys.platform.startswith('win32'):  # nocover
            # get the directory name for the current user
            c_users = dirname(userhome())
            userhome_dpath = join(c_users, username)
            if not exists(userhome_dpath):
                raise KeyError('Unknown user: {}'.format(username))
        else:
            import pwd
            try:
                pwent = pwd.getpwnam(username)
            except KeyError:  # nocover
                raise KeyError('Unknown user: {}'.format(username))
            userhome_dpath = pwent.pw_dir
    return userhome_dpath


def shrinkuser(path, home='~'):
    """
    Inverse of :func:`os.path.expanduser`.

    Args:
        path (str | PathLike): path in system file structure
        home (str): symbol used to replace the home path.
            Defaults to '~', but you might want to use '$HOME' or
            '%USERPROFILE%' instead.

    Returns:
        str: path - shortened path replacing the home directory with a symbol

    Example:
        >>> from ubelt.util_path import *  # NOQA
        >>> path = expanduser('~')
        >>> assert path != '~'
        >>> assert shrinkuser(path) == '~'
        >>> assert shrinkuser(path + '1') == path + '1'
        >>> assert shrinkuser(path + '/1') == join('~', '1')
        >>> assert shrinkuser(path + '/1', '$HOME') == join('$HOME', '1')
        >>> assert shrinkuser('.') == '.'
    """
    path = normpath(path)
    userhome_dpath = userhome()
    if path.startswith(userhome_dpath):
        if len(path) == len(userhome_dpath):
            path = home
        elif path[len(userhome_dpath)] == os.path.sep:
            path = home + path[len(userhome_dpath):]
    return path


def expandpath(path):
    """
    Shell-like environment variable and tilde path expansion.

    Args:
        path (str | PathLike): string representation of a path

    Returns:
        str : expanded path

    Example:
        >>> from ubelt.util_path import *  # NOQA
        >>> import ubelt as ub
        >>> assert normpath(ub.expandpath('~/foo')) == join(ub.userhome(), 'foo')
        >>> assert ub.expandpath('foo') == 'foo'
    """
    path = expanduser(path)
    path = expandvars(path)
    return path


def ensuredir(dpath, mode=0o1777, verbose=0, recreate=False):
    r"""
    Ensures that directory will exist. Creates new dir with sticky bits by
    default

    Args:
        dpath (str | PathLike | Tuple[str | PathLike]): dir to ensure. Can also
            be a tuple to send to join
        mode (int): octal mode of directory
        verbose (int): verbosity
        recreate (bool): if True removes the directory and
            all of its contents and creates a fresh new directory.
            USE CAREFULLY.

    Returns:
        str: path - the ensured directory

    SeeAlso:
        :func:`ubelt.Path.ensuredir`

    Note:
        This function is not thread-safe in Python2

    Example:
        >>> from ubelt.util_path import *  # NOQA
        >>> import ubelt as ub
        >>> cache_dpath = ub.Path.appdir('ubelt').ensuredir()
        >>> dpath = join(cache_dpath, 'ensuredir')
        >>> if exists(dpath):
        ...     os.rmdir(dpath)
        >>> assert not exists(dpath)
        >>> ub.ensuredir(dpath)
        >>> assert exists(dpath)
        >>> os.rmdir(dpath)
    """
    if isinstance(dpath, (list, tuple)):
        dpath = join(*dpath)

    if recreate:
        import ubelt as ub
        ub.delete(dpath, verbose=verbose)

    if not exists(dpath):
        if verbose:
            print('Ensuring directory (creating {!r})'.format(dpath))
        os.makedirs(normpath(dpath), mode=mode, exist_ok=True)
    else:
        if verbose:
            print('Ensuring directory (existing {!r})'.format(dpath))
    return dpath


class TempDir(object):
    """
    Context for creating and cleaning up temporary directories.

    Note:
        This class will be DEPRECATED. The exact deprecation version and
        mitigation plan has not yet been developed.

    Note:
        This exists because :class:`tempfile.TemporaryDirectory` was
        introduced in Python 3.2. Thus once ubelt no longer supports
        python 2.7, this class will be deprecated.

    Example:
        >>> from ubelt.util_path import *  # NOQA
        >>> with TempDir() as self:
        >>>     dpath = self.dpath
        >>>     assert exists(dpath)
        >>> assert not exists(dpath)

    Example:
        >>> from ubelt.util_path import *  # NOQA
        >>> self = TempDir()
        >>> dpath = self.ensure()
        >>> assert exists(dpath)
        >>> self.cleanup()
        >>> assert not exists(dpath)
    """
    def __init__(self):
        import ubelt as ub
        ub.schedule_deprecation(
            modname='ubelt',
            migration='Use tempfile instead', name='TempDir',
            type='class', deprecate='1.2.0', error='1.4.0',
            remove='1.5.0',
        )
        self.dpath = None

    def __del__(self):
        self.cleanup()

    def ensure(self):
        import tempfile
        if not self.dpath:
            self.dpath = tempfile.mkdtemp()
        return self.dpath

    def cleanup(self):
        if self.dpath:
            import shutil
            shutil.rmtree(self.dpath)
            self.dpath = None

    def start(self):
        self.ensure()
        return self

    def __enter__(self):
        return self.start()

    def __exit__(self, type_, value, trace):
        self.cleanup()


_PathBase = pathlib.WindowsPath if os.name == 'nt' else pathlib.PosixPath


class Path(_PathBase):
    """
    This class extends :class:`pathlib.Path` with extra functionality and
    convenience methods.

    New methods are designed to support chaining.

    In addition to new methods this class supports the addition (``+``)
    operator via which allows for better drop-in compatibility with code using
    existing string-based paths.

    Note:
        On windows this inherits from :class:`pathlib.WindowsPath`.

    New methods are

        * :py:meth:`ubelt.Path.ensuredir` - Like mkdir but with easier defaults.

        * :py:meth:`ubelt.Path.delete` - Previously pathlib could only remove one file at a time.

        * :py:meth:`ubelt.Path.copy` - Pathlib has no similar functionality.

        * :py:meth:`ubelt.Path.move` - Pathlib has no similar functionality.

        * :py:meth:`ubelt.Path.augment` - Unifies and extends disparate functionality across pathlib.

        * :py:meth:`ubelt.Path.expand` - Unifies existing environ and home expansion.

        * :py:meth:`ubelt.Path.ls` - Like iterdir, but more interactive.

        * :py:meth:`ubelt.Path.shrinkuser` - Python has no similar functionality.

        * :py:meth:`ubelt.Path.walk` - Pathlib had no similar functionality.

    New classmethods are

        * :py:meth:`ubelt.Path.appdir` - application directories

    Modified methods are

        * :py:meth:`ubelt.Path.touch` - returns self to support chaining

    Example:
        >>> # Ubelt extends pathlib functionality
        >>> import ubelt as ub
        >>> # Chain expansion and mkdir with cumbersome args.
        >>> dpath = ub.Path('~/.cache/ubelt/demo_path').expand().ensuredir()
        >>> fpath = dpath / 'text_file.txt'
        >>> # Augment is concise and chainable
        >>> aug_fpath = fpath.augment(stemsuffix='.aux', ext='.jpg').touch()
        >>> aug_dpath = dpath.augment(stemsuffix='demo_path2')
        >>> assert aug_fpath.read_text() == ''
        >>> fpath.write_text('text data')
        >>> assert aug_fpath.exists()
        >>> # Delete is akin to "rm -rf" and is also chainable.
        >>> assert not aug_fpath.delete().exists()
        >>> assert dpath.exists()
        >>> assert not dpath.delete().exists()
        >>> print(f'{str(fpath.shrinkuser()).replace(os.path.sep, "/")}')
        >>> print(f'{str(dpath.shrinkuser()).replace(os.path.sep, "/")}')
        >>> print(f'{str(aug_fpath.shrinkuser()).replace(os.path.sep, "/")}')
        >>> print(f'{str(aug_dpath.shrinkuser()).replace(os.path.sep, "/")}')
        ~/.cache/ubelt/demo_path/text_file.txt
        ~/.cache/ubelt/demo_path
        ~/.cache/ubelt/demo_path/text_file.aux.jpg
        ~/.cache/ubelt/demo_pathdemo_path2

    Inherited unmodified properties from :class:`pathlib.Path` are:

        * :py:data:`pathlib.PurePath.anchor`
        * :py:data:`pathlib.PurePath.name`
        * :py:data:`pathlib.PurePath.parts`
        * :py:data:`pathlib.PurePath.parts`
        * :py:data:`pathlib.PurePath.parent`
        * :py:data:`pathlib.PurePath.parents`
        * :py:data:`pathlib.PurePath.suffix`
        * :py:data:`pathlib.PurePath.suffixes`
        * :py:data:`pathlib.PurePath.stem`
        * :py:data:`pathlib.PurePath.drive`
        * :py:data:`pathlib.PurePath.root`

    Inherited unmodified classmethods from :class:`pathlib.Path` are:

        * :py:meth:`pathlib.Path.cwd`
        * :py:meth:`pathlib.Path.home`

    Inherited unmodified methods from :class:`pathlib.Path` are:

        * :py:meth:`pathlib.Path.samefile`
        * :py:meth:`pathlib.Path.iterdir`

        * :py:meth:`pathlib.Path.glob`
        * :py:meth:`pathlib.Path.rglob`

        * :py:meth:`pathlib.Path.resolve`

        * :py:meth:`pathlib.Path.lstat`
        * :py:meth:`pathlib.Path.stat`
        * :py:meth:`pathlib.Path.owner`
        * :py:meth:`pathlib.Path.group`

        * :py:meth:`pathlib.Path.open`
        * :py:meth:`pathlib.Path.read_bytes`
        * :py:meth:`pathlib.Path.read_text`
        * :py:meth:`pathlib.Path.write_bytes`
        * :py:meth:`pathlib.Path.write_text`
        * :py:meth:`pathlib.Path.readlink`

        * :py:meth:`pathlib.Path.mkdir` - we recommend :py:meth:`ubelt.Path.ensuredir` instead.

        * :py:meth:`pathlib.Path.chmod`
        * :py:meth:`pathlib.Path.lchmod`

        * :py:meth:`pathlib.Path.unlink`
        * :py:meth:`pathlib.Path.rmdir`

        * :py:meth:`pathlib.Path.rename`
        * :py:meth:`pathlib.Path.replace`

        * :py:meth:`pathlib.Path.symlink_to`
        * :py:meth:`pathlib.Path.hardlink_to`
        * :py:meth:`pathlib.Path.link_to` - deprecated

        * :py:meth:`pathlib.Path.exists`
        * :py:meth:`pathlib.Path.is_dir`
        * :py:meth:`pathlib.Path.is_file`
        * :py:meth:`pathlib.Path.is_mount`
        * :py:meth:`pathlib.Path.is_symlink`
        * :py:meth:`pathlib.Path.is_block_device`
        * :py:meth:`pathlib.Path.is_char_device`
        * :py:meth:`pathlib.Path.is_fifo`
        * :py:meth:`pathlib.Path.is_socket`

        * :py:meth:`pathlib.Path.expanduser` - we recommend :py:meth:`ubelt.Path.expand` instead.

        * :py:meth:`pathlib.PurePath.as_posix`
        * :py:meth:`pathlib.PurePath.as_uri`

        * :py:meth:`pathlib.PurePath.with_name` - we recommend :py:meth:`ubelt.Path.augment` instead.
        * :py:meth:`pathlib.PurePath.with_stem`  - we recommend :py:meth:`ubelt.Path.augment` instead.
        * :py:meth:`pathlib.PurePath.with_suffix` - we recommend :py:meth:`ubelt.Path.augment` instead.

        * :py:meth:`pathlib.PurePath.relative_to`

        * :py:meth:`pathlib.PurePath.joinpath`

        * :py:meth:`pathlib.PurePath.is_relative_to`
        * :py:meth:`pathlib.PurePath.is_absolute`
        * :py:meth:`pathlib.PurePath.is_reserved`

        * :py:meth:`pathlib.PurePath.match`
    """
    __slots__ = ()

    @classmethod
    def appdir(cls, appname=None, *args, type='cache'):
        """
        Returns a standard platform specific directory for an application to
        use as cache, config, or data.

        The default root location depends on the platform and is specified the
        the following table:

        TextArt:

                   | POSIX            | Windows        | MacOSX
            data   | $XDG_DATA_HOME   | %APPDATA%      | ~/Library/Application Support
            config | $XDG_CONFIG_HOME | %APPDATA%      | ~/Library/Application Support
            cache  | $XDG_CACHE_HOME  | %LOCALAPPDATA% | ~/Library/Caches


            If an environment variable is not specified the defaults are:
                APPDATA      = ~/AppData/Roaming
                LOCALAPPDATA = ~/AppData/Local

                XDG_DATA_HOME   = ~/.local/share
                XDG_CACHE_HOME  = ~/.cache
                XDG_CONFIG_HOME = ~/.config

        Args:
            appname (str | None):
                The name of the application.

            *args : optional subdirs

            type (str):
                the type of data the expected to be stored in this application
                directory. Valid options are 'cache', 'config', or 'data'.

        Returns:
            Path: a new path object for the specified application directory.

        SeeAlso:
            This provides functionality similar to the
            `appdirs <https://pypi.org/project/appdirs/>`_ -
            and
            `platformdirs <https://platformdirs.readthedocs.io/en/latest/api.html>`_ -
            packages.

        Example:
            >>> # xdoctest: +IGNORE_WANT
            >>> import ubelt as ub
            >>> print(ub.Path.appdir('ubelt', type='cache').shrinkuser())
            >>> print(ub.Path.appdir('ubelt', type='config').shrinkuser())
            >>> print(ub.Path.appdir('ubelt', type='data').shrinkuser())
            ~/.cache/ubelt
            ~/.config/ubelt
            ~/.local/share/ubelt
            >>> import pytest
            >>> with pytest.raises(KeyError):
            >>>     ub.Path.appdir('ubelt', type='other')

        Example:
            >>> # xdoctest: +IGNORE_WANT
            >>> import ubelt as ub
            >>> # Can now call appdir without any arguments
            >>> print(ub.Path.appdir().shrinkuser())
            ~/.cache
        """
        from ubelt import util_platform
        if type == 'cache':
            base = util_platform.platform_cache_dir()
        elif type == 'config':
            base = util_platform.platform_config_dir()
        elif type == 'data':
            base = util_platform.platform_data_dir()
        else:
            raise KeyError(type)

        if appname is None:
            return cls(base, *args)
        else:
            return cls(base, appname, *args)

    def augment(self, prefix='', stemsuffix='', ext=None, stem=None, dpath=None,
                tail='', relative=None, multidot=False, suffix=''):
        """
        Create a new path with a different extension, basename, directory,
        prefix, and/or suffix.

        See :func:`augpath` for more details.

        Args:
            prefix (str):
                Text placed in front of the stem. Defaults to ''.

            stemsuffix (str):
                Text placed between the stem and extension. Default to ''.
                Note: this is just called suffix in :func:`ub.augpath`.

            ext (str | None):
                If specified, replaces the extension

            stem (str | None):
                If specified, replaces the stem (i.e. basename without
                extension). Note: named base in :func:`augpath`.

            dpath (str | PathLike | None):
                If specified, replaces the specified "relative" directory,
                which by default is the parent directory.

            tail (str | None):
                If specified, appends this text to the extension.

            relative (str | PathLike | None):
                Replaces ``relative`` with ``dpath`` in ``path``.
                Has no effect if ``dpath`` is not specified.
                Defaults to the dirname of the input ``path``.
                *experimental* not currently implemented.

            multidot (bool): Allows extensions to contain
                multiple dots. Specifically, if False, everything after the
                last dot in the basename is the extension. If True, everything
                after the first dot in the basename is the extension.

            suffix (str):
                DEPRECAETD

        SeeAlso:
            :py:meth:`pathlib.Path.with_stem`
            :py:meth:`pathlib.Path.with_name`
            :py:meth:`pathlib.Path.with_suffix`

        Returns:
            Path: augmented path

        Note:
            NOTICE OF BACKWARDS INCOMPATABILITY.

            THE INITIAL RELEASE OF Path.augment suffered from an unfortunate
            variable naming decision that conflicts with pathlib.Path

            .. code:: python

            p = ub.Path('the.entire.fname.or.dname.is.the.name.exe')
            print(f'p     ={p}')
            print(f'p.name={p.name}')
            p = ub.Path('the.stem.ends.here.ext')
            print(f'p     ={p}')
            print(f'p.stem={p.stem}')
            p = ub.Path('only.the.last.dot.is.the.suffix')
            print(f'p       ={p}')
            print(f'p.suffix={p.suffix}')
            p = ub.Path('but.all.suffixes.can.be.recovered')
            print(f'p         ={p}')
            print(f'p.suffixes={p.suffixes}')

            p.name
            p.stem
            p.suffixes
            p.parts

        Example:
            >>> import ubelt as ub
            >>> path = ub.Path('foo.bar')
            >>> suffix = '_suff'
            >>> prefix = 'pref_'
            >>> ext = '.baz'
            >>> newpath = path.augment(prefix=prefix, stemsuffix=suffix, ext=ext, stem='bar')
            >>> print('newpath = {!r}'.format(newpath))
            newpath = Path('pref_bar_suff.baz')

        Example:
            >>> import ubelt as ub
            >>> path = ub.Path('foo.bar')
            >>> stemsuffix = '_suff'
            >>> prefix = 'pref_'
            >>> ext = '.baz'
            >>> newpath = path.augment(prefix=prefix, stemsuffix=stemsuffix, ext=ext, stem='bar')
            >>> print('newpath = {!r}'.format(newpath))

        Example:
            >>> # Compare our augpath(ext=...) versus pathlib with_suffix(...)
            >>> import ubelt as ub
            >>> cases = [
            >>>     ub.Path('no_ext'),
            >>>     ub.Path('one.ext'),
            >>>     ub.Path('double..dot'),
            >>>     ub.Path('two.many.cooks'),
            >>>     ub.Path('path.with.three.dots'),
            >>>     ub.Path('traildot.'),
            >>>     ub.Path('doubletraildot..'),
            >>>     ub.Path('.prefdot'),
            >>>     ub.Path('..doubleprefdot'),
            >>> ]
            >>> for path in cases:
            >>>     print('--')
            >>>     print('path = {}'.format(ub.repr2(path, nl=1)))
            >>>     ext = '.EXT'
            >>>     method_pathlib = path.with_suffix(ext)
            >>>     method_augment = path.augment(ext=ext)
            >>>     if method_pathlib == method_augment:
            >>>         print(ub.color_text('sagree', 'green'))
            >>>     else:
            >>>         print(ub.color_text('disagree', 'red'))
            >>>     print('path.with_suffix({}) = {}'.format(ext, ub.repr2(method_pathlib, nl=1)))
            >>>     print('path.augment(ext={}) = {}'.format(ext, ub.repr2(method_augment, nl=1)))
            >>>     print('--')
        """
        if suffix:  # nocover
            from ubelt.util_deprecate import schedule_deprecation
            schedule_deprecation(
                'ubelt', 'suffix', 'arg',
                deprecate='1.1.3', remove='2.0.0',
                migration='Use stemsuffix instead',
            )
            if not stemsuffix:
                stemsuffix = suffix
            import warnings
            warnings.warn(
                'DEVELOPER NOTICE: The ubelt.Path.augment function may '
                'experience a BACKWARDS INCOMPATIBLE update in the future '
                'having to do with the suffix argument to ub.Path.augment '
                'To avoid any issue use the ``ubelt.augment`` function '
                'instead for now. If you see this warning, please make an '
                'issue on https://github.com/Erotemic/ubelt/issues indicating '
                'that there are users of this function in the wild. If there '
                'are none, then this signature will be "fixed", but if anyone '
                'depends on this feature then we will continue to support it as '
                'is.'
            )

        aug = augpath(self, suffix=stemsuffix, prefix=prefix, ext=ext, base=stem,
                      dpath=dpath, relative=relative, multidot=multidot,
                      tail=tail)
        new = self.__class__(aug)
        return new

    def delete(self):
        """
        Removes a file or recursively removes a directory.
        If a path does not exist, then this is does nothing.

        SeeAlso:
            :func:`ubelt.delete`

        Returns:
            Path: reference to self

        Example:
            >>> import ubelt as ub
            >>> from os.path import join
            >>> base = ub.Path.appdir('ubelt', 'delete_test2')
            >>> dpath1 = (base / 'dir').ensuredir()
            >>> (base / 'dir' / 'subdir').ensuredir()
            >>> (base / 'dir' / 'to_remove1.txt').touch()
            >>> fpath1 = (base / 'dir' / 'subdir' / 'to_remove3.txt').touch()
            >>> fpath2 = (base / 'dir' / 'subdir' / 'to_remove2.txt').touch()
            >>> assert all(p.exists() for p in [dpath1, fpath1, fpath2])
            >>> fpath1.delete()
            >>> assert all(p.exists() for p in [dpath1, fpath2])
            >>> assert not fpath1.exists()
            >>> dpath1.delete()
            >>> assert not any(p.exists() for p in [dpath1, fpath1, fpath2])
        """
        util_io.delete(self)
        return self

    def ensuredir(self, mode=0o777):
        """
        Concise alias of ``self.mkdir(parents=True, exist_ok=True)``

        Returns:
            Path: returns itself

        Example:
            >>> import ubelt as ub
            >>> cache_dpath = ub.Path.appdir('ubelt').ensuredir()
            >>> dpath = ub.Path(join(cache_dpath, 'ensuredir'))
            >>> if dpath.exists():
            ...     os.rmdir(dpath)
            >>> assert not dpath.exists()
            >>> dpath.ensuredir()
            >>> assert dpath.exists()
            >>> dpath.rmdir()
            """
        self.mkdir(mode=mode, parents=True, exist_ok=True)
        return self

    def mkdir(self, mode=511, parents=False, exist_ok=False):
        """
        Create a new directory at this given path.

        Note:
            The ubelt variant is the same, except it returns the path as well.

        Args:
            mode (int) : perms
            parents (bool) : create parents
            exist_ok (bool) : fail if exists

        Returns:
            Path: returns itself
        """
        super().mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
        return self

    def expand(self):
        """
        Expands user tilde and environment variables.

        Concise alias of ``Path(os.path.expandvars(self.expanduser()))``

        Returns:
            Path: path with expanded environment variables and tildes

        Example:
            >>> import ubelt as ub
            >>> #home_v1 = ub.Path('$HOME').expand()
            >>> home_v2 = ub.Path('~/').expand()
            >>> assert isinstance(home_v2, ub.Path)
            >>> home_v3 = ub.Path.home()
            >>> #print('home_v1 = {!r}'.format(home_v1))
            >>> print('home_v2 = {!r}'.format(home_v2))
            >>> print('home_v3 = {!r}'.format(home_v3))
            >>> assert home_v3 == home_v2 # == home_v1
        """
        return self.expandvars().expanduser()

    def expandvars(self):
        """
        As discussed in [CPythonIssue21301]_, CPython won't be adding
        expandvars to pathlib. I think this is a mistake, so I added it in this
        extension.

        Returns:
            Path: path with expanded environment variables

        References:
            .. [CPythonIssue21301] https://bugs.python.org/issue21301
        """
        return self.__class__(os.path.expandvars(self))

    def ls(self):
        """
        A convenience function to list all paths in a directory.

        This is simply a wraper around iterdir that returns the results as a
        list instead of a generator. This is mainly for faster navigation in
        IPython. In production code `iterdir` should be used instead.

        Returns:
            List[Path]

        Example:
            >>> import ubelt as ub
            >>> self = ub.Path.appdir('ubelt/tests/ls')
            >>> (self / 'dir1').ensuredir()
            >>> (self / 'dir2').ensuredir()
            >>> (self / 'file1').touch()
            >>> (self / 'file2').touch()
            >>> (self / 'dir1/file3').touch()
            >>> (self / 'dir2/file4').touch()
            >>> children = self.ls()
            >>> assert isinstance(children, list)
            >>> print(ub.repr2(sorted([p.relative_to(self) for p in children])))
            [
                Path('dir1'),
                Path('dir2'),
                Path('file1'),
                Path('file2'),
            ]
        """
        return list(self.iterdir())

    def shrinkuser(self, home='~'):
        """
        Shrinks your home dir by replacing it with a tilde.

        This is the inverse of :func:`os.path.expanduser`.

        Args:
            home (str): symbol used to replace the home path.
                Defaults to '~', but you might want to use '$HOME' or
                '%USERPROFILE%' instead.

        Returns:
            Path: path - shortened path replacing the home directory with a symbol

        Example:
            >>> import ubelt as ub
            >>> path = ub.Path('~').expand()
            >>> assert str(path.shrinkuser()) == '~'
            >>> assert str(ub.Path((str(path) + '1')).shrinkuser()) == str(path) + '1'
            >>> assert str((path / '1').shrinkuser()) == join('~', '1')
            >>> assert str((path / '1').shrinkuser('$HOME')) == join('$HOME', '1')
            >>> assert str(ub.Path('.').shrinkuser()) == '.'
        """
        shrunk = shrinkuser(self, home)
        new = self.__class__(shrunk)
        return new

    def touch(self, mode=0o666, exist_ok=True):
        """
        Create this file with the given access mode, if it doesn't exist.

        Returns:
            Path: returns itself

        Note:
            The :func:`ubelt.util_io.touch` function currently has a slightly
            different implementation. This uses whatever the pathlib version
            is. This may change in the future.
        """
        # modify touch to return self
        # Note: util_io.touch is more expressive than standard python
        # touch, may want to use that instead.
        super().touch(mode=mode, exist_ok=exist_ok)
        return self

    def walk(self, topdown=True, onerror=None, followlinks=False):
        """
        A variant of :func:`os.walk` for pathlib

        Args:
            topdown (bool):
                if True starts yield nodes closer to the root first otherwise
                yield nodes closer to the leaves first.

            onerror (Callable[[OSError], None]):
                A function with one argument of type OSError. If the
                error is raised the walk is aborted, otherwise it continues.

            followlinks (bool):
                if True recurse into symbolic directory links

        Yields:
            Tuple['Path', List[str], List[str]]:
                the root path, directory names, and file names

        Example:
            >>> import ubelt as ub
            >>> self = ub.Path.appdir('ubelt/tests/ls')
            >>> (self / 'dir1').ensuredir()
            >>> (self / 'dir2').ensuredir()
            >>> (self / 'file1').touch()
            >>> (self / 'file2').touch()
            >>> (self / 'dir1/file3').touch()
            >>> (self / 'dir2/file4').touch()
            >>> subdirs = list(self.walk())
            >>> assert len(subdirs) == 3

        Example:
            >>> # Modified from the stdlib
            >>> import os
            >>> from os.path import join, getsize
            >>> import email
            >>> import ubelt as ub
            >>> base = ub.Path(email.__file__).parent
            >>> for root, dirs, files in base.walk():
            >>>     print(root, " consumes", end="")
            >>>     print(sum(getsize(join(root, name)) for name in files), end="")
            >>>     print("bytes in ", len(files), " non-directory files")
            >>>     if 'CVS' in dirs:
            >>>         dirs.remove('CVS')  # don't visit CVS directories
        """
        cls = self.__class__
        walker = os.walk(self, topdown=topdown, onerror=onerror,
                         followlinks=followlinks)
        for root, dnames, fnames in walker:
            yield (cls(root), dnames, fnames)

    def __add__(self, other):
        """
        Returns a new string starting with this fspath representation.

        Returns:
            str

        Allows ubelt.Path to be a better drop-in replacement when working with
        string-based paths.

        Note:
            It is not recommended to write new code that uses this behavior.
            This exists to make it easier to transition existing str-based
            paths to pathlib.

        Example:
            >>> import ubelt as ub
            >>> base = ub.Path('base')
            >>> base_ = ub.Path('base/')
            >>> base2 = ub.Path('base/2')
            >>> assert base + 'foo' == 'basefoo'
            >>> assert base_ + 'foo' == 'basefoo'
            >>> assert base2 + 'foo' == str(base2.augment(tail='foo'))
        """
        return os.fspath(self) + other

    def __radd__(self, other):
        """
        Returns a new string ending with this fspath representation.

        Returns:
            str

        Allows ubelt.Path to be a better drop-in replacement when working with
        string-based paths.

        Note:
            It is not recommended to write new code that uses this behavior.
            This exists to make it easier to transition existing str-based
            paths to pathlib.

        Example:
            >>> import ubelt as ub
            >>> base = ub.Path('base')
            >>> base_ = ub.Path('base/')
            >>> base2 = ub.Path('base/2')
            >>> assert 'foo' + base == 'foobase'
            >>> assert 'foo' + base_ == 'foobase'
            >>> assert 'foo' + base2 == str(base2.augment(dpath='foobase'))
        """
        return other + os.fspath(self)

    def endswith(self, suffix, *args):
        """
        Test if the fspath representation endswith a particular string

        Allows ubelt.Path to be a better drop-in replacement when working with
        string-based paths.

        Args:
            suffix (str | Tuple[str, ...]):
                One or more suffixes to test for

            *args:
                start (int): if specified begin testing at this position.
                end (int): if specified stop testing at this position.

        Returns:
            bool: True if any of the suffixes are matched.

        Example:
            >>> import ubelt as ub
            >>> base = ub.Path('base')
            >>> assert base.endswith('se')
            >>> assert not base.endswith('be')
            >>> # test start / stop cases
            >>> assert ub.Path('aabbccdd').endswith('cdd', 5)
            >>> assert not ub.Path('aabbccdd').endswith('cdd', 6)
            >>> assert ub.Path('aabbccdd').endswith('cdd', 5, 10)
            >>> assert not ub.Path('aabbccdd').endswith('cdd', 5, 7)
            >>> # test tuple case
            >>> assert ub.Path('aabbccdd').endswith(('foo', 'cdd'))
            >>> assert ub.Path('foo').endswith(('foo', 'cdd'))
            >>> assert not ub.Path('bar').endswith(('foo', 'cdd'))
        """
        return os.fspath(self).endswith(suffix, *args)

    def startswith(self, prefix, *args):
        """
        Test if the fspath representation startswith a particular string

        Allows ubelt.Path to be a better drop-in replacement when working with
        string-based paths.

        Args:
            prefix (str | Tuple[str, ...]):
                One or more prefixes to test for

            *args:
                start (int): if specified begin testing at this position.
                end (int): if specified stop testing at this position.

        Returns:
            bool: True if any of the prefixes are matched.

        Example:
            >>> import ubelt as ub
            >>> base = ub.Path('base')
            >>> assert base.startswith('base')
            >>> assert not base.startswith('all your')
            >>> # test start / stop cases
            >>> assert ub.Path('aabbccdd').startswith('aab', 0)
            >>> assert ub.Path('aabbccdd').startswith('aab', 0, 5)
            >>> assert not ub.Path('aabbccdd').startswith('aab', 1, 5)
            >>> assert not ub.Path('aabbccdd').startswith('aab', 0, 2)
            >>> # test tuple case
            >>> assert ub.Path('aabbccdd').startswith(('foo', 'aab'))
            >>> assert ub.Path('foo').startswith(('foo', 'aab'))
            >>> assert not ub.Path('bar').startswith(('foo', 'aab'))
        """
        return os.fspath(self).startswith(prefix, *args)

    # More shutil functionality
    # This is discussed in https://peps.python.org/pep-0428/#filesystem-modification

    def _request_copy_function(self, follow_file_symlinks=True,
                               follow_dir_symlinks=True, meta='stats'):
        """
        Get a copy_function based on specified capabilities
        """
        import shutil
        from functools import partial
        if meta is None:
            copy_function = partial(shutil.copyfile, follow_symlinks=follow_file_symlinks)
        elif meta == 'stats':
            copy_function = partial(shutil.copy2, follow_symlinks=follow_file_symlinks)
        elif meta == 'mode':
            copy_function = partial(shutil.copy, follow_symlinks=follow_file_symlinks)
        else:
            raise KeyError(meta)
        return copy_function

    def copy(self, dst, follow_file_symlinks=False, follow_dir_symlinks=False,
             meta='stats', overwrite=False):
        """
        Copy this file or directory to dst.

        By default files are never overwritten and symlinks are copied as-is.


        At a basic level (i.e. ignoring symlinks) for each path argument
        (``src`` and ``dst``) these can either be files, directories, or not
        exist. Given these three states, the following table summarizes how
        this function copies this path to its destination.

        TextArt:

            +----------+------------------------+------------------------+------------+
            | dst      | dir                    | file                   | no-exist   |
            +----------+                        |                        |            |
            | src      |                        |                        |            |
            +==========+========================+========================+============+
            | dir      | error-or-overwrite-dst | error                  | dst        |
            +----------+------------------------+------------------------+------------+
            | file     | dst / src.name         | error-or-overwrite-dst | dst        |
            +----------+------------------------+------------------------+------------+
            | no-exist | error                  | error                  | error      |
            +----------+------------------------+------------------------+------------+

        In general, the contents of src will be the contents of dst, except for
        the one case where a file is copied into an existing directory. In this
        case the name is used to construct a fully qualified destination.

        Ignore:
            # Enumerate cases
            rows = [
                {'src': 'no-exist', 'dst': 'no-exist', 'result': 'error'},
                {'src': 'no-exist', 'dst': 'file',     'result': 'error'},
                {'src': 'no-exist', 'dst': 'dir',      'result': 'error'},

                {'src': 'file', 'dst': 'no-exist', 'result': 'dst'},
                {'src': 'file', 'dst': 'dir',      'result': 'dst / src.name'},
                {'src': 'file', 'dst': 'file',     'result': 'error-or-overwrite-dst'},

                {'src': 'dir', 'dst': 'no-exist', 'result': 'dst'},
                {'src': 'dir', 'dst': 'dir',      'result': 'error-or-overwrite-dst'},
                {'src': 'dir', 'dst': 'file',     'result': 'error'},
            ]
            import pandas as pd
            df = pd.DataFrame(rows)
            piv = df.pivot(['src'], ['dst'], 'result')
            print(piv.to_markdown(tablefmt="presto"))
            print(piv.to_markdown(tablefmt="pretty", index=True))
            print(piv.to_markdown(tablefmt="grid", index=True))

        Args:
            dst (str | PathLike):
                if `src` is a file and `dst` does not exist, copies this to `dst`
                if `src` is a file and `dst` is a directory, copies this to `dst / src.name`

                if `src` is a directory and `dst` does not exist, copies this to `dst`
                if `src` is a directory and `dst` is a directory, errors unless
                overwrite is True, in which case, copies this to `dst` and
                overwrites anything conflicting path.

            follow_file_symlinks (bool):
                If True and src is a link, the link will be resolved before
                it is copied (i.e. the data is duplicated), otherwise just
                the link itself will be copied.

            follow_dir_symlinks (bool):
                if True when src is a directory and contains symlinks to
                other directories, the contents of the linked data are
                copied, otherwise when False only the link itself is
                copied.

            meta (str | None):
                Indicates what metadata bits to copy. This can be 'stats' which
                tries to copy all metadata (i.e. like :py:func:`shutil.copy2`),
                'mode' which copies just the permission bits (i.e. like
                :py:func:`shutil.copy`), or None, which ignores all metadata
                (i.e.  like :py:func:`shutil.copyfile`).

            overwrite (bool):
                if False, and target file exists, this will raise an error,
                otherwise the file will be overwritten.

        Returns:
            Path: where the path was actually copied to

        Note:
            This is implemented with a combination of :func:`shutil.copy`,
            :func:`shutil.copy2`, and :func:`shutil.copytree`, but the The
            defaults and behavior here are noticably different (and hopefully
            safer and more intuitive).

        Example:
            >>> import ubelt as ub
            >>> root = ub.Path.appdir('ubelt', 'tests', 'path', 'copy').delete().ensuredir()
            >>> paths = {}
            >>> dpath = (root / 'orig').ensuredir()
            >>> clone0 = (root / 'dst_is_explicit').ensuredir()
            >>> clone1 = (root / 'dst_is_parent').ensuredir()
            >>> paths['fpath'] = (dpath / 'file0.txt').touch()
            >>> paths['empty_dpath'] = (dpath / 'empty_dpath').ensuredir()
            >>> paths['nested_dpath'] = (dpath / 'nested_dpath').ensuredir()
            >>> (dpath / 'nested_dpath/d0').ensuredir()
            >>> (dpath / 'nested_dpath/d0/f1.txt').touch()
            >>> (dpath / 'nested_dpath/d0/f2.txt').touch()
            >>> print('paths = {}'.format(ub.repr2(paths, nl=1)))
            >>> assert all(p.exists() for p in paths.values())
            >>> paths['fpath'].copy(clone0 / 'file0.txt')
            >>> paths['fpath'].copy(clone1)
            >>> paths['empty_dpath'].copy(clone0 / 'empty_dpath')
            >>> paths['empty_dpath'].copy((clone1 / 'empty_dpath_alt').ensuredir(), overwrite=True)
            >>> paths['nested_dpath'].copy(clone0 / 'nested_dpath')
            >>> paths['nested_dpath'].copy((clone1 / 'nested_dpath_alt').ensuredir(), overwrite=True)

        Ignore:
            import xdev
            xdev.tree_repr(dpath)
            xdev.tree_repr(clone0, max_files=10)
            xdev.tree_repr(clone1, max_files=10)
        """
        warnings.warn('The ub.Path.copy function is experimental and may change, '
                      'in corner cases. Primary cases seem stable.')
        import shutil
        copy_function = self._request_copy_function(
            follow_file_symlinks=follow_file_symlinks,
            follow_dir_symlinks=follow_dir_symlinks, meta=meta)
        if self.is_dir():
            if sys.version_info[0:2] < (3, 8):  # nocover
                copytree = _compat_copytree
            else:
                copytree = shutil.copytree
            dst = copytree(
                self, dst, copy_function=copy_function,
                symlinks=not follow_dir_symlinks, dirs_exist_ok=overwrite)
        elif self.is_file():
            if not overwrite:
                dst = Path(dst)
                if dst.is_dir():
                    real_dst = dst / self.name
                else:
                    real_dst = dst
                if real_dst.exists():
                    raise FileExistsError('Cannot overwrite existing file unless overwrite=True')
            dst = copy_function(self, dst)
        else:
            raise Exception
        return Path(dst)

    def move(self, dst, follow_file_symlinks=False, follow_dir_symlinks=False,
             meta='stats'):
        """
        Move a file from one location to another, or recursively move a
        directory from one location to another.

        This method will refuse to overwrite anything, and there is currently
        no overwrite option for technical reasons. This may change in the
        future.

        Args:
            dst (str | PathLike):
                A non-existing path where this file will be moved.

            follow_file_symlinks (bool):
                If True and src is a link, the link will be resolved before
                it is copied (i.e. the data is duplicated), otherwise just
                the link itself will be copied.

            follow_dir_symlinks (bool):
                if True when src is a directory and contains symlinks to
                other directories, the contents of the linked data are
                copied, otherwise when False only the link itself is
                copied.

            meta (str | None):
                Indicates what metadata bits to copy. This can be 'stats' which
                tries to copy all metadata (i.e. like shutil.copy2), 'mode'
                which copies just the permission bits (i.e. like shutil.copy),
                or None, which ignores all metadata (i.e.  like
                shutil.copyfile).

        Note:
            This method will refuse to overwrite anything.

            This is implemented via :func:`shutil.move`, which depends heavily
            on :func:`os.rename` semantics. For this reason, this function
            error if it would overwrite any data. If you want an overwriting
            variant of move we recommend you either either copy the data, and
            then delete the original (potentially inefficient), or use
            :func:`shutil.move` directly if you know how :func:`os.rename`
            works on your system.

        Returns:
            Path: where the path was actually moved to

        Example:
            >>> import ubelt as ub
            >>> dpath = ub.Path.appdir('ubelt', 'tests', 'path', 'move').delete().ensuredir()
            >>> paths = {}
            >>> paths['dpath0'] = (dpath / 'dpath0').ensuredir()
            >>> paths['dpath00'] = (dpath / 'dpath0' / 'sub0').ensuredir()
            >>> paths['fpath000'] = (dpath / 'dpath0' / 'sub0' / 'f0.txt').touch()
            >>> paths['fpath001'] = (dpath / 'dpath0' / 'sub0' / 'f1.txt').touch()
            >>> paths['dpath01'] = (dpath / 'dpath0' / 'sub1').ensuredir()
            >>> print('paths = {}'.format(ub.repr2(paths, nl=1)))
            >>> assert all(p.exists() for p in paths.values())
            >>> # xdev.tree_repr(dpath, max_files=10)
            >>> paths['dpath0'].move(dpath / 'dpath1')
            >>> # xdev.tree_repr(dpath, max_files=10)

        Ignore:
            import xdev
            xdev.tree_repr(dpath)
        """
        warnings.warn('The ub.Path.move function is experimental and may change! '
                      'Do not rely on this behavior yet!')

        # Behave more like POSIX move to avoid potential confusing behavior
        if exists(dst):
            raise FileExistsError(
                'Moves are only allowed to locations that dont exist')
        # if os.path.isdir(dst):
        #     with os.scandir(dst) as itr:
        #         is_empty = next(itr, None) is None
        #     if not is_empty:
        #         raise IOError(f'Cannot move {self!r} to {dst!r}. '
        #                       'Directory not empty')

        import shutil
        copy_function = self._request_copy_function(
            follow_file_symlinks=follow_file_symlinks,
            follow_dir_symlinks=follow_dir_symlinks, meta=meta)
        real_dst = shutil.move(self, dst, copy_function=copy_function)
        return Path(real_dst)

if sys.version_info[0:2] < (3, 8):  # nocover

    # Vendor in a nearly modern copytree for Python 3.6 and 3.7
    def _compat_copytree(src, dst, symlinks=False, ignore=None,
                         copy_function=None, ignore_dangling_symlinks=False,
                         dirs_exist_ok=False):
        """
        A vendored shutil.copytree for older pythons based on the 3.10
        implementation
        """
        # import stat
        from shutil import Error, copystat, copy2, copy
        with os.scandir(src) as itr:
            entries = list(itr)

        if ignore is not None:
            ignored_names = ignore(os.fspath(src), [x.name for x in entries])
        else:
            ignored_names = set()

        os.makedirs(dst, exist_ok=dirs_exist_ok)
        errors = []
        use_srcentry = copy_function is copy2 or copy_function is copy

        for srcentry in entries:
            if srcentry.name in ignored_names:
                continue
            srcname = os.path.join(src, srcentry.name)
            dstname = os.path.join(dst, srcentry.name)
            srcobj = srcentry if use_srcentry else srcname
            try:
                is_symlink = srcentry.is_symlink()
                if is_symlink and os.name == 'nt':
                    # Special check for directory junctions, which appear as
                    # symlinks but we want to recurse.
                    # Not available on 3.6, use our impl instead
                    # lstat = srcentry.stat(follow_symlinks=False)
                    # if lstat.st_reparse_tag == stat.IO_REPARSE_TAG_MOUNT_POINT:
                    #   is_symlink = False
                    from ubelt._win32_links import _win32_is_junction
                    if _win32_is_junction(srcentry):
                        is_symlink = False
                if is_symlink:
                    linkto = os.readlink(srcname)
                    if symlinks:
                        # We can't just leave it to `copy_function` because legacy
                        # code with a custom `copy_function` may rely on copytree
                        # doing the right thing.
                        os.symlink(linkto, dstname)
                        copystat(srcobj, dstname, follow_symlinks=not symlinks)
                    else:
                        # ignore dangling symlink if the flag is on
                        if not os.path.exists(linkto) and ignore_dangling_symlinks:
                            continue
                        # otherwise let the copy occur. copy2 will raise an error
                        if srcentry.is_dir():
                            _compat_copytree(srcobj, dstname, symlinks, ignore,
                                             copy_function,
                                             dirs_exist_ok=dirs_exist_ok)
                        else:
                            copy_function(srcobj, dstname)
                elif srcentry.is_dir():
                    _compat_copytree(srcobj, dstname, symlinks, ignore,
                                     copy_function,
                                     dirs_exist_ok=dirs_exist_ok)
                else:
                    # Will raise a SpecialFileError for unsupported file types
                    copy_function(srcobj, dstname)
            # catch the Error from the recursive copytree so that we can
            # continue with other files
            except Error as err:
                errors.extend(err.args[0])
            except OSError as why:
                errors.append((srcname, dstname, str(why)))
        try:
            copystat(src, dst)
        except OSError as why:
            # Copying file access times may fail on Windows
            if getattr(why, 'winerror', None) is None:
                errors.append((src, dst, str(why)))
        if errors:
            raise Error(errors)
        return dst



from typing import Type
from typing import Tuple
from typing import Union
from typing import Callable


def urepr(data: object, **kwargs) -> str:
    ...


def repr2(data, **kwargs):
    ...


class FormatterExtensions:

    def __init__(self) -> None:
        ...

    def register(self, key: Union[Type, Tuple[Type], str]) -> Callable:
        ...

    def lookup(self, data):
        ...



from typing import Mapping
from typing import Sequence
from typing import Union
from typing import Iterable
from typing import Callable
from typing import Iterator
from typing import List
from _typeshed import Incomplete
from collections.abc import Generator
from typing import Any, TypeVar

VT = TypeVar("VT")
T = TypeVar("T")
KT = TypeVar("KT")


class chunks:
    legacy: Incomplete
    remainder: Incomplete
    items: Incomplete
    total: Incomplete
    nchunks: Incomplete
    chunksize: Incomplete
    bordermode: Incomplete

    def __init__(self,
                 items,
                 chunksize: Incomplete | None = ...,
                 nchunks: Incomplete | None = ...,
                 total: Incomplete | None = ...,
                 bordermode: str = ...,
                 legacy: bool = ...) -> None:
        ...

    def __len__(self):
        ...

    def __iter__(self):
        ...

    @staticmethod
    def noborder(items, chunksize) -> Generator[Any, None, None]:
        ...

    @staticmethod
    def cycle(items, chunksize) -> Generator[Any, None, None]:
        ...

    @staticmethod
    def replicate(items, chunksize) -> Generator[Any, None, None]:
        ...


def iterable(obj: object, strok: bool = ...) -> bool:
    ...


def take(items: Union[Sequence[VT], Mapping[KT, VT]],
         indices: Iterable[Union[int, KT]],
         default: Any = ...) -> Generator[VT, None, None]:
    ...


def compress(items: Iterable[Any], flags: Iterable[bool]) -> Iterable[Any]:
    ...


def flatten(nested: Iterable[Iterable[Any]]) -> Iterable[Any]:
    ...


def unique(items: Iterable[T],
           key: Callable[[T], Any] = ...) -> Generator[T, None, None]:
    ...


def argunique(items: Sequence[VT],
              key: Callable[[VT], Any] = ...) -> Iterator[int]:
    ...


def unique_flags(items: Sequence[VT],
                 key: Union[Callable[[VT], Any], None] = ...) -> List[bool]:
    ...


def boolmask(indices: List[int], maxval: int = ...) -> List[bool]:
    ...


def iter_window(iterable: Iterable[T],
                size: int = ...,
                step: int = ...,
                wrap: bool = ...) -> Iterable[T]:
    ...


def allsame(iterable: Iterable[T], eq: Callable[[T, T], bool] = ...) -> bool:
    ...


def argsort(indexable: Union[Iterable[VT], Mapping[KT, VT]],
            key: Union[Callable[[VT], VT], None] = ...,
            reverse: bool = ...) -> List[int] | List[KT]:
    ...


def argmax(indexable: Union[Iterable[VT], Mapping[KT, VT]],
           key: Callable[[VT], Any] = ...) -> int | KT:
    ...


def argmin(indexable: Union[Iterable[VT], Mapping[KT, VT]],
           key: Callable[[VT], VT] = ...) -> int | KT:
    ...


def peek(iterable: Iterable[T], default: T = ...) -> T:
    ...



from typing import Union


def schedule_deprecation(modname: str,
                         name: str = ...,
                         type: str = ...,
                         migration: str = ...,
                         deprecate: Union[str, None] = ...,
                         error: Union[str, None] = ...,
                         remove: Union[str, None] = ...,
                         warncls: type = ...,
                         stacklevel: int = ...) -> str:
    ...



from typing import Iterable
from typing import Type
from typing import Dict
from typing import Callable
from typing import Union
from typing import List
from typing import Optional
from ubelt.util_const import NoParamType
from typing import Mapping
from typing import Set
import sys
from collections import OrderedDict, defaultdict
from collections.abc import Generator
from typing import Any, TypeVar

VT = TypeVar("VT")
T = TypeVar("T")
KT = TypeVar("KT")
odict = OrderedDict
ddict = defaultdict
DictBase = OrderedDict if sys.version_info[0:2] <= (3, 6) else dict


def dzip(items1: Iterable[KT],
         items2: Iterable[VT],
         cls: Type[dict] = ...) -> Dict[KT, VT]:
    ...


def group_items(
        items: Iterable[VT],
        key: Union[Iterable[KT], Callable[[VT], KT]]) -> dict[KT, List[VT]]:
    ...


def dict_hist(items: Iterable[T],
              weights: Iterable[float] = ...,
              ordered: bool = ...,
              labels: Iterable[T] = ...) -> dict[T, int]:
    ...


def find_duplicates(items: Iterable[T],
                    k: int = ...,
                    key: Callable[[T], Any] = ...) -> dict[T, List[int]]:
    ...


def dict_subset(dict_: Dict[KT, VT],
                keys: Iterable[KT],
                default: Union[Optional[object], NoParamType] = ...,
                cls: Type[Dict] = ...) -> Dict[KT, VT]:
    ...


def dict_union(*args: List[Dict]) -> Dict | OrderedDict:
    ...


def dict_diff(
    *args: List[Union[Dict[KT, VT], Iterable[KT]]]
) -> Dict[KT, VT] | OrderedDict[KT, VT]:
    ...


def dict_isect(
    *args: List[Union[Dict[KT, VT], Iterable[KT]]]
) -> Dict[KT, VT] | OrderedDict[KT, VT]:
    ...


def map_values(func: Union[Callable[[VT], T], Mapping[VT, T]],
               dict_: Dict[KT, VT],
               cls: Union[type, None] = ...) -> Dict[KT, T]:
    ...


map_vals = map_values


def map_keys(func: Union[Callable[[KT], T], Mapping[KT, T]],
             dict_: Dict[KT, VT],
             cls: Union[type, None] = ...) -> Dict[T, VT]:
    ...


def sorted_values(dict_: Dict[KT, VT],
                  key: Union[Callable[[VT], Any], None] = ...,
                  reverse: bool = ...,
                  cls: type = ...) -> OrderedDict[KT, VT]:
    ...


sorted_vals = sorted_values


def sorted_keys(dict_: Dict[KT, VT],
                key: Union[Callable[[KT], Any], None] = ...,
                reverse: bool = ...,
                cls: type = ...) -> OrderedDict[KT, VT]:
    ...


def invert_dict(
        dict_: Dict[KT, VT],
        unique_vals: bool = ...,
        cls: Union[type, None] = ...) -> Dict[VT, KT] | Dict[VT, Set[KT]]:
    ...


def named_product(
        _: Union[Dict[str, List[VT]], None] = ...,
        **basis: Dict[str, List[VT]]) -> Generator[Dict[str, VT], None, None]:
    ...


def varied_values(longform: List[Dict[KT, VT]],
                  min_variations: int = ...,
                  default: Union[VT, NoParamType] = ...) -> Dict[KT, List[VT]]:
    ...


class SetDict(dict):

    def copy(self):
        ...

    def __or__(self, other):
        ...

    def __and__(self, other):
        ...

    def __sub__(self, other):
        ...

    def __xor__(self, other):
        ...

    def __ror__(self, other):
        ...

    def __rand__(self, other):
        ...

    def __rsub__(self, other):
        ...

    def __rxor__(self, other):
        ...

    def __ior__(self, other):
        ...

    def __iand__(self, other):
        ...

    def __isub__(self, other):
        ...

    def __ixor__(self, other):
        ...

    def union(self, *others, cls: type = ...) -> dict:
        ...

    def intersection(self, *others, cls: type = ...) -> dict:
        ...

    def difference(self, *others, cls: type = ...) -> dict:
        ...

    def symmetric_difference(self, *others, cls: type = ...) -> dict:
        ...


sdict = SetDict


class UDict(SetDict):

    def subdict(self,
                keys: Iterable[KT],
                default: Union[Optional[object], NoParamType] = ...):
        ...

    def take(
        self,
        keys: Iterable[KT],
        default: Union[Optional[object], NoParamType] = ...
    ) -> Generator[VT, None, None]:
        ...

    def invert(self,
               unique_vals: bool = ...) -> Dict[VT, KT] | Dict[VT, Set[KT]]:
        ...

    def map_keys(
            self, func: Union[Callable[[VT], T], Mapping[VT,
                                                         T]]) -> Dict[KT, T]:
        ...

    def map_values(
            self, func: Union[Callable[[VT], T], Mapping[VT,
                                                         T]]) -> Dict[KT, T]:
        ...

    def sorted_keys(self,
                    key: Union[Callable[[KT], Any], None] = ...,
                    reverse: bool = ...) -> OrderedDict[KT, VT]:
        ...

    def sorted_values(self,
                      key: Union[Callable[[VT], Any], None] = ...,
                      reverse: bool = ...) -> OrderedDict[KT, VT]:
        ...

    def peek_key(self, default: Union[KT, NoParamType] = ...) -> KT:
        ...

    def peek_value(self, default: Union[VT, NoParamType] = ...) -> VT:
        ...


class AutoDict(UDict):

    def __getitem__(self, key):
        ...

    def to_dict(self) -> dict:
        ...


AutoOrderedDict = AutoDict
udict = UDict



from typing import Union
from os import PathLike
import concurrent.futures
from _typeshed import Incomplete


class DownloadManager:
    pool: Incomplete
    download_root: Incomplete
    cache: Incomplete
    dl_func: Incomplete

    def __init__(self,
                 download_root: Incomplete | None = ...,
                 mode: str = ...,
                 max_workers: Incomplete | None = ...,
                 cache: bool = ...) -> None:
        ...

    def submit(self,
               url: Union[str, PathLike],
               dst: Union[str, None] = ...,
               hash_prefix: Union[str, None] = ...,
               hasher: str = ...) -> concurrent.futures.Future:
        ...

    def as_completed(self,
                     prog: Incomplete | None = ...,
                     desc: Incomplete | None = ...,
                     verbose: int = ...):
        ...

    def shutdown(self) -> None:
        ...

    def __iter__(self):
        ...

    def __len__(self):
        ...



"""
Simple ways to interact with the commandline without defining a full blown
CLI. These are usually used for developer hacks. Any real interface should
probably be defined using :py:mod:`argparse`, :py:mod:`click`, or
:py:mod:`scriptconfig`. Be sure to ignore unknown arguments if you use them in
conjunction with these functions.

The :func:`argflag` function checks if a boolean ``--flag`` style CLI argument
exists on the command line.

The :func:`argval` function returns the value of a ``--key=value`` style CLI
argument.
"""
import sys
from ubelt import util_const

__all__ = ['argval', 'argflag']


def argval(key, default=util_const.NoParam, argv=None):
    """
    Get the value of a keyword argument specified on the command line.

    Values can be specified as ``<key> <value>`` or ``<key>=<value>``

    The use-case for this function is to add hidden command line feature where
    a developer can pass in a special value. This can be used to prototype a
    command line interface, provide an easter egg, or add some other command
    line parsing that wont be exposed in CLI help docs.

    Args:
        key (str | Tuple[str, ...]):
            string or tuple of strings. Each key should be prefixed with two
            hyphens (i.e. ``--``)

        default (T | NoParamType, default=NoParam):
            a value to return if not specified.

        argv (Optional[List[str]], default=None):
            uses ``sys.argv`` if unspecified

    Returns:
        str | T:
            value - the value specified after the key. It they key is specified
            multiple times, then the first value is returned.

    TODO:
        - [x] Can we handle the case where the value is a list of long paths? - No
        - [ ] Should we default the first or last specified instance of the flag.

    CommandLine:
        xdoctest -m ubelt.util_arg argval:0
        xdoctest -m ubelt.util_arg argval:0 --devval
        xdoctest -m ubelt.util_arg argval:0 --devval=1
        xdoctest -m ubelt.util_arg argval:0 --devval=2
        xdoctest -m ubelt.util_arg argval:0 --devval 3
        xdoctest -m ubelt.util_arg argval:0 --devval "4 5 6"

    Example:
        >>> # Everyday usage of this function might look like this where
        >>> import ubelt as ub
        >>> # grab a key/value pair if is given on the command line
        >>> value = ub.argval('--devval', default='1')
        >>> print('Checking if the hidden CLI key/value pair is given')
        >>> if value != '1':
        >>>     print(ub.color_text(
        >>>         'A hidden developer secret: {!r}'.format(value), 'yellow'))
        >>> print('Pass the hidden CLI key/value pair to see a secret message')

    Example:
        >>> import ubelt as ub
        >>> argv = ['--ans', '42', '--quest=the grail', '--ans=6', '--bad']
        >>> assert ub.argval('--spam', argv=argv) == ub.NoParam
        >>> assert ub.argval('--quest', argv=argv) == 'the grail'
        >>> assert ub.argval('--ans', argv=argv) == '42'
        >>> assert ub.argval('--bad', argv=argv) == ub.NoParam
        >>> assert ub.argval(('--bad', '--bar'), argv=argv) == ub.NoParam

    Example:
        >>> # Test fix for GH Issue #41
        >>> import ubelt as ub
        >>> argv = ['--path=/path/with/k=3']
        >>> ub.argval('--path', argv=argv) == '/path/with/k=3'
    """
    if argv is None:  # nocover
        argv = sys.argv

    keys = [key] if isinstance(key, str) else key
    n_max = len(argv) - 1
    for argx, item in enumerate(argv):
        for key_ in keys:
            if item == key_:
                if argx < n_max:
                    value = argv[argx + 1]
                    return value
            elif item.startswith(key_ + '='):
                value = '='.join(item.split('=')[1:])
                return value
    value = default
    return value


def argflag(key, argv=None):
    """
    Determines if a key is specified on the command line.

    This is a functional alternative to ``key in sys.argv``, but it also allows
    for multiple aliases of the same flag to be specified.

    Args:
        key (str | Tuple[str, ...]):
            string or tuple of strings. Each key should be prefixed with two
            hyphens (i.e. ``--``).

        argv (List[str], default=None): overrides ``sys.argv`` if specified

    Returns:
        bool: flag - True if the key (or any of the keys) was specified

    CommandLine:
        xdoctest -m ubelt.util_arg argflag:0
        xdoctest -m ubelt.util_arg argflag:0 --devflag
        xdoctest -m ubelt.util_arg argflag:0 -df
        xdoctest -m ubelt.util_arg argflag:0 --devflag2
        xdoctest -m ubelt.util_arg argflag:0 -df2

    Example:
        >>> # Everyday usage of this function might look like this
        >>> import ubelt as ub
        >>> # Check if either of these strings are in sys.argv
        >>> flag = ub.argflag(('-df', '--devflag'))
        >>> if flag:
        >>>     print(ub.color_text(
        >>>         'A hidden developer flag was given!', 'blue'))
        >>> print('Pass the hidden CLI flag to see a secret message')

    Example:
        >>> import ubelt as ub
        >>> argv = ['--spam', '--eggs', 'foo']
        >>> assert ub.argflag('--eggs', argv=argv) is True
        >>> assert ub.argflag('--ans', argv=argv) is False
        >>> assert ub.argflag('foo', argv=argv) is True
        >>> assert ub.argflag(('bar', '--spam'), argv=argv) is True
    """
    if argv is None:  # nocover
        argv = sys.argv
    keys = [key] if isinstance(key, str) else key
    flag = any(k in argv for k in keys)
    return flag



class NoParamType:

    def __new__(cls):
        ...

    def __reduce__(self):
        ...

    def __copy__(self):
        ...

    def __deepcopy__(self, memo):
        ...

    def __bool__(self):
        ...


NoParam: NoParamType



"""
Abstractions for working with zipfiles and archives

This may be renamed to util_archive in the future.

The :func:`ubelt.split_archive` works with paths that reference a file inside
of an archive (e.g. a zipfile). It splits it into two parts, the full path to
the archive and then the path to the file inside of the archive. By convention
these are separated with either a pathsep or a colon.

The :func:`ubelt.zopen` works to open a file that lives inside of an archive
without the user needing to worry about extracting it first. When possible it
will read it directly from the archive, but in some cases it may extract it to
a temporary directory first.
"""
import io
import os
from os.path import exists, join
from ubelt.util_mixins import NiceRepr

__all__ = ['zopen', 'split_archive']


def split_archive(fpath, ext='.zip'):
    """
    If fpath specifies a file inside a zipfile, it breaks it into two parts the
    path to the zipfile and the internal path in the zipfile.

    Example:
        >>> split_archive('/a/b/foo.txt')
        >>> split_archive('/a/b/foo.zip/bar.txt')
        >>> split_archive('/a/b/foo.zip/baz/biz.zip/bar.py')
        >>> split_archive('archive.zip')
        >>> import ubelt as ub
        >>> split_archive(ub.Path('/a/b/foo.zip/baz/biz.zip/bar.py'))
        >>> split_archive('/a/b/foo.zip/baz.pt/bar.zip/bar.zip', '.pt')

    TODO:
        Fix got/want for win32

        (None, None)
        ('/a/b/foo.zip', 'bar.txt')
        ('/a/b/foo.zip/baz/biz.zip', 'bar.py')
        ('archive.zip', None)
        ('/a/b/foo.zip/baz/biz.zip', 'bar.py')
        ('/a/b/foo.zip/baz.pt', 'bar.zip/bar.zip')
    """
    import re
    fpath = os.fspath(fpath)
    # fpath = os.fspath(fpath)
    pat = '({}[{}/:])'.format(re.escape(ext), re.escape(os.path.sep))
    # pat = r'(\'' + ext + '[' + re.escape(os.path.sep) + '/:])'
    parts = re.split(pat, fpath, flags=re.IGNORECASE)
    if len(parts) > 2:
        archivepath = ''.join(parts[:-1])[:-1]
        internal = parts[-1]
    elif len(parts) == 1:
        archivepath = parts[0]
        if not archivepath.endswith(ext):
            archivepath = None
        internal = None
    else:  # nocover
        raise AssertionError('impossible state')
    return archivepath, internal


class zopen(NiceRepr):
    """
    An abstraction of the normal :func:`open` function that can also handle
    reading data directly inside of zipfiles.

    This is a file-object like interface [FileObj] --- i.e. it supports the
    read and write methods to an underlying resource.

    Can open a file normally or open a file within a zip file (readonly).
    Tries to read from memory only, but will extract to a tempfile if necessary.

    Just treat the zipfile like a directory,
    e.g. /path/to/myzip.zip/compressed/path.txt OR?
    e.g. /path/to/myzip.zip:compressed/path.txt

    References:
        .. [FileObj] https://docs.python.org/3/glossary.html#term-file-object

    TODO:
        - [ ] Fast way to open a base zipfile, query what is inside, and
              then choose a file to further zopen (and passing along the same
              open zipfile reference maybe?).
        - [ ] Write mode in some restricted setting?

    Args:
        fpath (str | PathLike):
            path to a file, or a special path that denotes both a
            path to a zipfile and a path to a archived file inside of
            the zipfile.

        mode (str):
            Currently only "r" - readonly mode is supported

        seekable (bool):
            If True, attempts to force "seekability" of the underlying
            file-object, for compressed files this will first extract
            the file to a temporary location on disk.  If False, any underlying
            compressed file will be opened directly which may result in the
            object being non-seekable.

        ext (str):
            The extension of the zipfile. Modify this is a non-standard
            extension is used (e.g. for torch packages).

    Example:
        >>> from ubelt.util_zip import *  # NOQA
        >>> import pickle
        >>> import ubelt as ub
        >>> dpath = ub.Path.appdir('ubelt/tests/util_zip').ensuredir()
        >>> dpath = ub.Path(dpath)
        >>> data_fpath = dpath / 'test.pkl'
        >>> data = {'demo': 'data'}
        >>> with open(str(data_fpath), 'wb') as file:
        >>>     pickle.dump(data, file)
        >>> # Write data
        >>> import zipfile
        >>> zip_fpath = dpath / 'test_zip.archive'
        >>> stl_w_zfile = zipfile.ZipFile(os.fspath(zip_fpath), mode='w')
        >>> stl_w_zfile.write(os.fspath(data_fpath), os.fspath(data_fpath.relative_to(dpath)))
        >>> stl_w_zfile.close()
        >>> stl_r_zfile = zipfile.ZipFile(os.fspath(zip_fpath), mode='r')
        >>> stl_r_zfile.namelist()
        >>> stl_r_zfile.close()
        >>> # Test zopen
        >>> self = zopen(zip_fpath / 'test.pkl', mode='rb', ext='.archive')
        >>> print(self._split_archive())
        >>> print(self.namelist())
        >>> self.close()
        >>> self = zopen(zip_fpath / 'test.pkl', mode='rb', ext='.archive')
        >>> recon1 = pickle.loads(self.read())
        >>> self.close()
        >>> self = zopen(zip_fpath / 'test.pkl', mode='rb', ext='.archive')
        >>> recon2 = pickle.load(self)
        >>> self.close()
        >>> assert recon1 == recon2
        >>> assert recon1 is not recon2

    Example:
        >>> # Test we can load json data from a zipfile
        >>> from ubelt.util_zip import *  # NOQA
        >>> import ubelt as ub
        >>> import json
        >>> import zipfile
        >>> dpath = ub.Path.appdir('ubelt/tests/util_zip').ensuredir()
        >>> infopath = join(dpath, 'info.json')
        >>> ub.writeto(infopath, '{"x": "1"}')
        >>> zippath = join(dpath, 'infozip.zip')
        >>> internal = 'folder/info.json'
        >>> with zipfile.ZipFile(zippath, 'w') as myzip:
        >>>     myzip.write(infopath, internal)
        >>> fpath = zippath + '/' + internal
        >>> # Test context manager
        >>> with zopen(fpath, 'r') as self:
        >>>     info2 = json.load(self)
        >>>     assert info2['x'] == '1'
        >>> # Test outside of context manager
        >>> self = zopen(fpath, 'r')
        >>> print(self._split_archive())
        >>> info2 = json.load(self)
        >>> assert info2['x'] == '1'
        >>> # Test nice repr (with zfile)
        >>> print('self = {!r}'.format(self))
        >>> self.close()

    Example:
        >>> # Coverage tests --- move to unit-test
        >>> from ubelt.util_zip import *  # NOQA
        >>> import ubelt as ub
        >>> import json
        >>> import zipfile
        >>> dpath = ub.Path.appdir('ubelt/tests/util_zip').ensuredir()
        >>> textpath = join(dpath, 'seekable_test.txt')
        >>> text = chr(10).join(['line{}'.format(i) for i in range(10)])
        >>> ub.writeto(textpath, text)
        >>> zippath = join(dpath, 'seekable_test.zip')
        >>> internal = 'folder/seekable_test.txt'
        >>> with zipfile.ZipFile(zippath, 'w') as myzip:
        >>>     myzip.write(textpath, internal)
        >>> ub.delete(textpath)
        >>> fpath = zippath + '/' + internal
        >>> # Test seekable
        >>> self_seekable = zopen(fpath, 'r', seekable=True)
        >>> assert self_seekable.seekable()
        >>> self_seekable.seek(8)
        >>> assert self_seekable.readline() == 'ne1' + chr(10)
        >>> assert self_seekable.readline() == 'line2' + chr(10)
        >>> self_seekable.seek(8)
        >>> assert self_seekable.readline() == 'ne1' + chr(10)
        >>> assert self_seekable.readline() == 'line2' + chr(10)
        >>> # Test non-seekable?
        >>> # Sometimes non-seekable files are still seekable
        >>> maybe_seekable = zopen(fpath, 'r', seekable=False)
        >>> if maybe_seekable.seekable():
        >>>     maybe_seekable.seek(8)
        >>>     assert maybe_seekable.readline() == 'ne1' + chr(10)
        >>>     assert maybe_seekable.readline() == 'line2' + chr(10)
        >>>     maybe_seekable.seek(8)
        >>>     assert maybe_seekable.readline() == 'ne1' + chr(10)
        >>>     assert maybe_seekable.readline() == 'line2' + chr(10)


    Example:
        >>> # More coverage tests --- move to unit-test
        >>> from ubelt.util_zip import *  # NOQA
        >>> import ubelt as ub
        >>> import pytest
        >>> dpath = ub.Path.appdir('ubelt/tests/util_zip').ensuredir()
        >>> with pytest.raises(OSError):
        >>>     self = zopen('', 'r')
        >>> # Test open non-zip exsting file
        >>> existing_fpath = join(dpath, 'exists.json')
        >>> ub.writeto(existing_fpath, '{"x": "1"}')
        >>> self = zopen(existing_fpath, 'r')
        >>> assert self.read() == '{"x": "1"}'
        >>> # Test dir
        >>> dir(self)
        >>> # Test nice
        >>> print(self)
        >>> print('self = {!r}'.format(self))
        >>> self.close()
        >>> # Test open non-zip non-existing file
        >>> nonexisting_fpath = join(dpath, 'does-not-exist.txt')
        >>> ub.delete(nonexisting_fpath)
        >>> with pytest.raises(OSError):
        >>>     self = zopen(nonexisting_fpath, 'r')
        >>> with pytest.raises(NotImplementedError):
        >>>     self = zopen(nonexisting_fpath, 'w')
        >>> # Test nice-repr
        >>> self = zopen(existing_fpath, 'r')
        >>> print('self = {!r}'.format(self))
        >>> # pathological
        >>> self = zopen(existing_fpath, 'r')
        >>> self._handle = None
        >>> dir(self)
    """
    def __init__(self, fpath, mode='r', seekable=False, ext='.zip'):
        self.fpath = fpath
        self.ext = ext
        self.name = fpath
        self.mode = mode
        self._seekable = seekable
        self._zfpath = None  # points to the base zipfile (if appropriate)
        self._temp_dpath = None  # for temporary extraction
        self._zfile_read = None  # underlying opened zipfile object
        # The _handle pointer should be a file-like object that this zopen
        # object impersonate, by forwarding most every getattr call to it.
        self._handle = None
        self._open()

    @property
    def zfile(self):
        """
        Access the underlying archive file
        """
        if self._zfile_read is None:
            import zipfile
            archivefile, internal = self._split_archive()
            myzip = zipfile.ZipFile(archivefile, 'r')
            self._zfile_read = myzip
        return self._zfile_read

    def namelist(self):
        """
        Lists the contents of this zipfile
        """
        myzip = self.zfile
        namelist = myzip.namelist()
        return namelist

    def __nice__(self):
        if self._zfpath is None:
            return 'handle={}, mode={}'.format(str(self._handle), self.mode)
        else:
            return 'handle={} in zipfpath={}, mode={}'.format(self._handle, self._zfpath, self.mode)

    def __getattr__(self, key):
        # Expose attributes of wrapped handle
        if hasattr(self._handle, key):
            assert self._handle is not self
            return getattr(self._handle, key)
        raise AttributeError(key)

    def __dir__(self):
        # Expose attributes of wrapped handle
        zopen_attributes = {
            'namelist',
            'zfile',
        }
        keyset = set(dir(super(zopen, self)))
        keyset.update(set(self.__dict__.keys()))
        if self._handle is not None:
            keyset.update(set(dir(self._handle)))
        return sorted(keyset | zopen_attributes)

    def _cleanup(self):
        # print('self._cleanup = {!r}'.format(self._cleanup))
        if self._handle is not None:
            if not getattr(self, 'closed', True):
                closemethod = getattr(self, 'close', None)
                if closemethod is not None:  # nocover
                    closemethod()
                closemethod = None
        self._handle = None
        if self._temp_dpath and exists(self._temp_dpath):
            # os.unlink(self._temp_dpath)
            import ubelt as ub
            ub.delete(self._temp_dpath)

    def __del__(self):
        self._cleanup()

    def _split_archive(self):
        archivefile, internal = split_archive(self.fpath, self.ext)
        return archivefile, internal

    def _open(self):
        """
        This logic sets the "_handle" to the appropriate backend object
        such that zopen can behave like a standard IO object.

        In read-only mode:
            * If fpath is a normal file, _handle is the standard `open` object
            * If fpath is a seekable zipfile, _handle is an IOWrapper pointing
                to the internal data
            * If fpath is a non-seekable zipfile, the data is extracted behind
                the scenes and a standard `open` object to the extracted file
                is given.

        In write mode:
            * NotImpelemented
        """
        if 'r' not in self.mode:
            raise NotImplementedError('Only read mode is supported for now')
        _handle = None
        fpath = os.fspath(self.fpath)
        if exists(fpath):
            _handle = open(fpath, self.mode)
        elif self.ext + '/' in fpath or self.ext + os.path.sep in fpath:
            archivefile, internal = self._split_archive()
            myzip = self.zfile
            if self._seekable:
                import tempfile
                # If we need data to be seekable, then we must extract it to a
                # temporary file first.
                self._temp_dpath = tempfile.mkdtemp(prefix='zopen_')
                temp_fpath = join(self._temp_dpath, internal)
                myzip.extract(internal, self._temp_dpath)
                _handle = open(temp_fpath, self.mode)
            else:
                # Try to load data directly from the zipfile
                _handle = myzip.open(internal, 'r')
                if self.mode == 'rb':
                    data = _handle.read()
                    _handle = io.BytesIO(data)
                elif self.mode == 'r':
                    # FIXME: does not always work. handle seems to be closed
                    # too soon in the case util.zopen(module.__file__).read()
                    _handle = io.TextIOWrapper(_handle)
                else:
                    raise KeyError(self.mode)
                self._zfpath = archivefile
        if _handle is None:
            raise IOError('file {!r} does not exist'.format(fpath))
        self._handle = _handle

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()



from typing import Tuple
from typing import Union
from typing import Callable
from ubelt.util_const import NoParamType
from typing import List
from os import PathLike
from _typeshed import Incomplete
from typing import TypeVar

Hasher = TypeVar("Hasher")
HASH_VERSION: int
DEFAULT_ALPHABET: Incomplete


def b(s):
    ...


DEFAULT_HASHER: Incomplete


class _Hashers:
    algos: Incomplete
    aliases: Incomplete

    def __init__(self) -> None:
        ...

    def available(self):
        ...

    def __contains__(self, key):
        ...

    def lookup(self, hasher):
        ...


class HashableExtensions:
    keyed_extensions: Incomplete
    iterable_checks: Incomplete

    def __init__(self) -> None:
        ...

    def register(self, hash_types: Union[type, Tuple[type]]) -> Callable:
        ...

    def lookup(self, data):
        ...

    def add_iterable_check(self, func):
        ...


class _HashTracer:
    sequence: Incomplete

    def __init__(self) -> None:
        ...

    def update(self, bytes) -> None:
        ...

    def hexdigest(self):
        ...


def hash_data(data: object,
              hasher: Union[str, Hasher, NoParamType] = ...,
              base: Union[List[str], str, NoParamType] = ...,
              types: bool = ...,
              convert: bool = ...,
              extensions: HashableExtensions = ...) -> str:
    ...


def hash_file(fpath: PathLike,
              blocksize: int = ...,
              stride: int = ...,
              maxbytes: Union[int, None] = ...,
              hasher: Union[str, Hasher, NoParamType] = ...,
              base: Union[List[str], str, NoParamType] = ...):
    ...



from typing import Union
from os import PathLike
from types import ModuleType
from typing import List
from typing import Tuple
from _typeshed import Incomplete


class PythonPathContext:
    dpath: Incomplete
    index: Incomplete

    def __init__(self, dpath, index: int = ...) -> None:
        ...

    def __enter__(self) -> None:
        ...

    def __exit__(self, type, value, trace) -> None:
        ...


def import_module_from_path(modpath: Union[str, PathLike],
                            index: int = ...) -> ModuleType:
    ...


def import_module_from_name(modname: str) -> ModuleType:
    ...


def modname_to_modpath(
        modname: str,
        hide_init: bool = ...,
        hide_main: bool = ...,
        sys_path: Union[None, List[Union[str, PathLike]]] = ...) -> str | None:
    ...


def normalize_modpath(modpath: Union[str, PathLike],
                      hide_init: bool = ...,
                      hide_main: bool = ...) -> str | PathLike:
    ...


def modpath_to_modname(modpath: str,
                       hide_init: bool = ...,
                       hide_main: bool = ...,
                       check: bool = ...,
                       relativeto: str = ...) -> str:
    ...


def split_modpath(modpath: str, check: bool = ...) -> Tuple[str, str]:
    ...


def is_modname_importable(modname: str,
                          sys_path: list = ...,
                          exclude: list = ...) -> bool:
    ...



from _typeshed import Incomplete
from collections.abc import MutableSet, Sequence

SLICE_ALL: Incomplete


def is_iterable(obj):
    ...


class OrderedSet(MutableSet, Sequence):
    items: Incomplete
    map: Incomplete

    def __init__(self, iterable: Incomplete | None = ...) -> None:
        ...

    def __len__(self):
        ...

    def __getitem__(self, index):
        ...

    def copy(self):
        ...

    def __contains__(self, key):
        ...

    def add(self, key):
        ...

    append: Incomplete

    def update(self, sequence):
        ...

    def index(self, key):
        ...

    get_loc: Incomplete
    get_indexer: Incomplete

    def pop(self):
        ...

    def discard(self, key) -> None:
        ...

    def clear(self) -> None:
        ...

    def __iter__(self):
        ...

    def __reversed__(self):
        ...

    def __eq__(self, other):
        ...

    def union(self, *sets):
        ...

    def __and__(self, other):
        ...

    def intersection(self, *sets):
        ...

    def difference(self, *sets):
        ...

    def issubset(self, other):
        ...

    def issuperset(self, other):
        ...

    def symmetric_difference(self, other):
        ...

    def difference_update(self, *sets) -> None:
        ...

    def intersection_update(self, other) -> None:
        ...

    def symmetric_difference_update(self, other) -> None:
        ...


oset = OrderedSet



from typing import Optional
import io
from typing import Union
from os import PathLike
from ubelt.util_const import NoParamType
from typing import Dict
import datetime
from typing import TypeVar

Hasher = TypeVar("Hasher")


def download(url: str,
             fpath: Optional[Union[str, PathLike, io.BytesIO]] = ...,
             dpath: Optional[PathLike] = ...,
             fname: Optional[str] = ...,
             appname: str = ...,
             hash_prefix: Union[None, str] = ...,
             hasher: Union[str, Hasher] = ...,
             chunksize: int = ...,
             verbose: Union[int, bool] = ...,
             timeout: Union[float, NoParamType] = ...,
             progkw: Union[Dict, NoParamType] = ...) -> str | PathLike:
    ...


def grabdata(url: str,
             fpath: Optional[Union[str, PathLike]] = ...,
             dpath: Optional[Union[str, PathLike]] = ...,
             fname: Optional[str] = ...,
             redo: bool = ...,
             verbose: int = ...,
             appname: str = ...,
             hash_prefix: Union[None, str] = ...,
             hasher: Union[str, Hasher] = ...,
             expires: Union[str, int, datetime.datetime] = ...,
             **download_kw) -> str | PathLike:
    ...



from typing import List
from typing import Union


def indent(text: str, prefix: str = ...) -> str:
    ...


def codeblock(text: str) -> str:
    ...


def paragraph(text: str) -> str:
    ...


def hzcat(args: List[str], sep: str = ...):
    ...


def ensure_unicode(text: Union[str, bytes]) -> str:
    ...



from typing import Union
from os import PathLike
import io


def writeto(fpath: Union[str, PathLike],
            to_write: str,
            aslines: bool = ...,
            verbose: bool = ...) -> None:
    ...


def readfrom(fpath: Union[str, PathLike],
             aslines: bool = ...,
             errors: str = ...,
             verbose: bool = ...) -> str:
    ...


def touch(fpath: Union[str, PathLike],
          mode: int = ...,
          dir_fd: io.IOBase = ...,
          verbose: int = ...,
          **kwargs) -> str:
    ...


def delete(path: Union[str, PathLike], verbose: bool = ...) -> None:
    ...



import concurrent.futures
from typing import Callable
from typing import Union
from typing import List
import concurrent.futures
from _typeshed import Incomplete
from collections.abc import Generator
from typing import Any


class SerialFuture(concurrent.futures.Future):
    func: Incomplete
    args: Incomplete
    kw: Incomplete

    def __init__(self, func, *args, **kw) -> None:
        ...

    def set_result(self, result) -> None:
        ...


class SerialExecutor:
    max_workers: int

    def __enter__(self):
        ...

    def __exit__(self, ex_type, ex_value, tb) -> None:
        ...

    def submit(self, func, *args, **kw) -> concurrent.futures.Future:
        ...

    def shutdown(self) -> None:
        ...

    def map(self, fn: Callable[..., Any], *iterables,
            **kwargs) -> Generator[Any, None, None]:
        ...


class Executor:
    backend: Incomplete

    def __init__(self, mode: str = ..., max_workers: int = ...) -> None:
        ...

    def __enter__(self):
        ...

    def __exit__(self, ex_type, ex_value, tb):
        ...

    def submit(self, func, *args, **kw) -> concurrent.futures.Future:
        ...

    def shutdown(self):
        ...

    def map(self, fn, *iterables, **kwargs):
        ...


class JobPool:
    executor: Incomplete
    jobs: Incomplete

    def __init__(self, mode: str = ..., max_workers: int = ...) -> None:
        ...

    def __len__(self):
        ...

    def submit(self, func: Callable[..., Any], *args,
               **kwargs) -> concurrent.futures.Future:
        ...

    def shutdown(self):
        ...

    def __enter__(self):
        ...

    def __exit__(self, a, b, c) -> None:
        ...

    def as_completed(
        self,
        timeout: Union[float, None] = ...,
        desc: Union[str, None] = ...,
        progkw: Union[dict, None] = ...
    ) -> Generator[concurrent.futures.Future, None, None]:
        ...

    def join(self, **kwargs) -> List[Any]:
        ...

    def __iter__(self):
        ...



import datetime
from typing import Union
from _typeshed import Incomplete


def timestamp(datetime: Union[datetime.datetime, datetime.date, None] = ...,
              precision: int = ...,
              default_timezone: Union[str, datetime.timezone] = ...,
              allow_dateutil: bool = ...) -> str:
    ...


def timeparse(stamp: str,
              default_timezone: str = ...,
              allow_dateutil: bool = ...) -> datetime.datetime:
    ...


class Timer:
    label: Incomplete
    verbose: Incomplete
    newline: Incomplete
    tstart: int
    elapsed: int
    write: Incomplete
    flush: Incomplete
    ns: Incomplete

    def __init__(self,
                 label: str = ...,
                 verbose: Incomplete | None = ...,
                 newline: bool = ...,
                 ns: bool = ...) -> None:
        ...

    def tic(self):
        ...

    def toc(self):
        ...

    def __enter__(self):
        ...

    def __exit__(self, ex_type, ex_value, trace):
        ...



from typing import Union
from os import PathLike
from typing import Tuple
from typing import List
from typing import Callable
from _typeshed import Incomplete
from collections.abc import Generator


def augpath(path: Union[str, PathLike],
            suffix: str = ...,
            prefix: str = ...,
            ext: Union[str, None] = ...,
            tail: Union[str, None] = ...,
            base: Union[str, None] = ...,
            dpath: Union[str, PathLike, None] = ...,
            relative: Union[str, PathLike, None] = ...,
            multidot: bool = ...) -> str:
    ...


def userhome(username: Union[str, None] = ...) -> str:
    ...


def shrinkuser(path: Union[str, PathLike], home: str = ...) -> str:
    ...


def expandpath(path: Union[str, PathLike]) -> str:
    ...


def ensuredir(dpath: Union[str, PathLike, Tuple[Union[str, PathLike]]],
              mode: int = ...,
              verbose: int = ...,
              recreate: bool = ...) -> str:
    ...


class TempDir:
    dpath: Incomplete

    def __init__(self) -> None:
        ...

    def __del__(self) -> None:
        ...

    def ensure(self):
        ...

    def cleanup(self) -> None:
        ...

    def start(self):
        ...

    def __enter__(self):
        ...

    def __exit__(self, type_, value, trace) -> None:
        ...


class Path:

    @classmethod
    def appdir(cls,
               appname: Union[str, None] = ...,
               *args,
               type: str = ...) -> 'Path':
        ...

    def augment(self,
                prefix: str = ...,
                stemsuffix: str = ...,
                ext: Union[str, None] = ...,
                stem: Union[str, None] = ...,
                dpath: Union[str, PathLike, None] = ...,
                tail: Union[str, None] = ...,
                relative: Union[str, PathLike, None] = ...,
                multidot: bool = ...,
                suffix: str = ...) -> 'Path':
        ...

    def delete(self) -> 'Path':
        ...

    def ensuredir(self, mode: int = ...) -> 'Path':
        ...

    def mkdir(self,
              mode: int = ...,
              parents: bool = ...,
              exist_ok: bool = ...) -> 'Path':
        ...

    def expand(self) -> 'Path':
        ...

    def expandvars(self) -> 'Path':
        ...

    def ls(self) -> List[Path]:
        ...

    def shrinkuser(self, home: str = ...) -> 'Path':
        ...

    def touch(self, mode: int = ..., exist_ok: bool = ...) -> 'Path':
        ...

    def walk(
        self,
        topdown: bool = ...,
        onerror: Callable[[OSError], None] = ...,
        followlinks: bool = ...
    ) -> Generator[Tuple['Path', List[str], List[str]], None, None]:
        ...

    def __add__(self, other) -> str:
        ...

    def __radd__(self, other) -> str:
        ...

    def endswith(self, suffix: Union[str, Tuple[str, ...]], *args) -> bool:
        ...

    def startswith(self, prefix: Union[str, Tuple[str, ...]], *args) -> bool:
        ...

    def copy(self,
             dst: Union[str, PathLike],
             follow_file_symlinks: bool = ...,
             follow_dir_symlinks: bool = ...,
             meta: Union[str, None] = ...,
             overwrite: bool = ...) -> 'Path':
        ...

    def move(self,
             dst: Union[str, PathLike],
             follow_file_symlinks: bool = ...,
             follow_dir_symlinks: bool = ...,
             meta: Union[str, None] = ...) -> 'Path':
        ...



from typing import Union
from os import PathLike
from typing import Callable
from _typeshed import Incomplete
from collections.abc import Generator
from typing import Any


class Cacher:
    VERBOSE: int
    FORCE_DISABLE: bool
    dpath: Incomplete
    fname: Incomplete
    depends: Incomplete
    cfgstr: Incomplete
    verbose: Incomplete
    ext: Incomplete
    meta: Incomplete
    enabled: Incomplete
    protocol: Incomplete
    hasher: Incomplete
    log: Incomplete
    backend: Incomplete

    def __init__(self,
                 fname,
                 depends: Incomplete | None = ...,
                 dpath: Incomplete | None = ...,
                 appname: str = ...,
                 ext: str = ...,
                 meta: Incomplete | None = ...,
                 verbose: Incomplete | None = ...,
                 enabled: bool = ...,
                 log: Incomplete | None = ...,
                 hasher: str = ...,
                 protocol: int = ...,
                 cfgstr: Incomplete | None = ...,
                 backend: str = ...) -> None:
        ...

    @property
    def fpath(self):
        ...

    def get_fpath(self, cfgstr: Union[str, None] = ...) -> str | PathLike:
        ...

    def exists(self, cfgstr: Union[str, None] = ...) -> bool:
        ...

    def existing_versions(self) -> Generator[str, None, None]:
        ...

    def clear(self, cfgstr: Union[str, None] = ...) -> None:
        ...

    def tryload(self,
                cfgstr: Union[str, None] = ...,
                on_error: str = ...) -> None | object:
        ...

    def load(self, cfgstr: Union[str, None] = ...) -> object:
        ...

    def save(self, data: object, cfgstr: Union[str, None] = ...) -> None:
        ...

    def ensure(self, func: Callable, *args, **kwargs):
        ...

    def __call__(self, func: Callable):
        ...


class CacheStamp:
    cacher: Incomplete
    product: Incomplete
    hasher: Incomplete
    expires: Incomplete
    hash_prefix: Incomplete

    def __init__(self,
                 fname,
                 dpath,
                 cfgstr: Incomplete | None = ...,
                 product: Incomplete | None = ...,
                 hasher: str = ...,
                 verbose: Incomplete | None = ...,
                 enabled: bool = ...,
                 depends: Incomplete | None = ...,
                 meta: Incomplete | None = ...,
                 hash_prefix: Incomplete | None = ...,
                 expires: Incomplete | None = ...,
                 ext: str = ...) -> None:
        ...

    @property
    def fpath(self):
        ...

    def clear(self):
        ...

    def expired(self, cfgstr: Any = ..., product: Any = ...) -> bool | str:
        ...

    def renew(self,
              cfgstr: Incomplete | None = ...,
              product: Incomplete | None = ...) -> None | dict:
        ...



from typing import List
from typing import Tuple
from typing import Dict
from typing import Union
from _typeshed import Incomplete
from collections.abc import Generator
from typing import Any


class IndexableWalker(Generator):
    data: Incomplete
    dict_cls: Incomplete
    list_cls: Incomplete
    indexable_cls: Incomplete

    def __init__(self, data, dict_cls=..., list_cls=...) -> None:
        ...

    def __iter__(self) -> Generator[Tuple[List, Any], Any, Any]:
        ...

    def __next__(self) -> Any:
        ...

    def send(self, arg) -> None:
        ...

    def throw(self,
              type: Incomplete | None = ...,
              value: Incomplete | None = ...,
              traceback: Incomplete | None = ...) -> None:
        ...

    def __setitem__(self, path: List, value: Any) -> None:
        ...

    def __getitem__(self, path: List) -> Any:
        ...

    def __delitem__(self, path: List) -> None:
        ...

    def allclose(self,
                 other: IndexableWalker,
                 rel_tol: float = ...,
                 abs_tol: float = ...,
                 return_info: bool = ...) -> bool | Tuple[bool, Dict]:
        ...


def indexable_allclose(items1: Union[dict, list, tuple],
                       items2: Union[dict, list, tuple],
                       rel_tol: float = ...,
                       abs_tol: float = ...,
                       return_info: bool = ...) -> bool | Tuple[bool, Dict]:
    ...



"""
This module defines :data:`ub.NoParam`. This is a robust sentinel value that
can act like ``None`` when None might be a valid value. The value of
:data:`NoParam` is robust to reloading, pickling, and copying (i.e. ``var is
ub.NoParam`` will return ``True`` after these operations).

Use cases that demonstrate the value of :data:`NoParam` can be found in
:mod:`ubelt.util_dict`, where it simplifies the implementation of methods that
behave like :meth:`dict.get`.

Example:
    >>> import ubelt as ub
    >>> def func(a=ub.NoParam):
    >>>     if a is ub.NoParam:
    >>>         print('no param specified')
    >>>     else:
    >>>         print('a = {}'.format(a))
    >>> func()
    no param specified
    >>> func(a=None)
    a = None
    >>> func(a=1)
    a = 1
    >>> # note: typically it is bad practice to use NoParam as an actual
    >>> # (non-default) parameter. It goes against the sprit of the idea.
    >>> func(a=ub.NoParam)
    no param specified
"""

__all__ = ['NoParam']


class NoParamType(object):
    r"""
    Class used to define :data:`NoParam`, a sentinel that acts like None when
    None might be a valid value. The value of :data:`NoParam` is robust to
    reloading, pickling, and copying. See [SO_41048643]_ for more details.

    However, try to never assign this value to a persistent variable.  Use this
    class sparingly.

    References:
        .. [SO_41048643]: http://stackoverflow.com/questions/41048643/a-second-none

    Example:
        >>> # Use case
        >>> # Imagine you need a function with a default argument, but you need to
        >>> # distinguish between cases where the user called it without specifying
        >>> # a default, versus the user specifying None. For instance, imagine you
        >>> # are writing the code for a dictionary ``get(key, default)``.
        >>> #
        >>> # You want the user to distinguish between the user calling it with
        >>> # None and the user not calling it at all.
        >>> #
        >>> # So you can't write it like this because you can't distinguish between
        >>> # the user passing default as None, or not passing a default at all.
        >>> def get(self, key, default=None):
        >>>     if default is None:
        >>>         ...  # What do?!
        >>> #
        >>> # You could write it like this, which is long and annoying
        >>> def get(self, key, *args, **kw):
        >>>     try:
        >>>         return self[key]
        >>>     except KeyError as ke:
        >>>         if len(args) > 0:
        >>>             return args[0]
        >>>         elif 'default' in kw:
        >>>             return kw['default']
        >>>         else:
        >>>             raise
        >>> #
        >>> # Instead write it like this, which is short and nice
        >>> from ubelt import NoParam
        >>> def get(self, key, default=NoParam):
        >>>     try:
        >>>         return self[key]
        >>>     except KeyError:
        >>>         if default is NoParam:
        >>>             raise
        >>>         return default
        >>> #
        >>> # setup some data
        >>> self = {}
        >>> key = 'spam'
        >>> #
        >>> # If the key is not in the dictionary, raise a KeyError
        >>> import pytest
        >>> with pytest.raises(KeyError):
        >>>     get(self, key)
        >>> #
        >>> # If the key is not in the dictionary, return ``default``
        >>> get(self, key, None)  # with positional args
        >>> get(self, key, default=None)  # with keyword args

    Example:
        >>> import ubelt as ub
        >>> from ubelt import util_const
        >>> from ubelt.util_const import NoParamType, NoParam
        >>> import pickle
        >>> import copy
        >>> id_ = id(NoParam)
        >>> versions = {
        ... 'ub.util_const.NoParam': ub.util_const.NoParam,
        ... 'NoParam': NoParam,
        ... 'NoParamType()': NoParamType(),
        ... 'ub.NoParam': ub.NoParam,
        ... 'copy': copy.copy(NoParam),
        ... 'deepcopy': copy.deepcopy(NoParam),
        ... 'pickle': pickle.loads(pickle.dumps(NoParam))
        ... }
        >>> print(versions)
        >>> assert all(id(v) == id_ for v in versions.values())
        >>> from importlib import reload
        >>> reload(util_const)
        >>> assert id(util_const.NoParam) == id_
        >>> assert all(id(v) == id_ for v in versions.values())
        >>> assert str(NoParam) == repr(NoParam)
        >>> assert not any(v for v in versions.values())
        >>> assert all(not v for v in versions.values())
        >>> assert all(not bool(v) for v in versions.values())
    """
    def __new__(cls):
        return NoParam
    def __reduce__(self):
        return (NoParamType, ())
    def __copy__(self):
        return NoParam
    def __deepcopy__(self, memo):
        return NoParam
    def __str__(cls):
        return 'NoParam'
    def __repr__(cls):
        return 'NoParam'
    def __bool__(self):
        # Ensure NoParam is Falsey
        return False

# Backwards compat
_NoParamType = NoParamType

# Create the only instance of NoParamType that should ever exist
try:
    # If the module is reloaded (via imp.reload), globals() will contain
    # NoParam. This skips the code that would instantiate a second object
    NoParam  # pragma: no cover
    # Note: it is possible to hack around this via
    # >>> del util_const.NoParam
    # >>> imp.reload(util_const)
except NameError:  # pragma: no cover
    # When the module is first loaded, globals() will not contain NoParam. A
    # NameError will be thrown, causing the first instance of NoParam to be
    # instantiated.
    NoParam = object.__new__(NoParamType)  # type: NoParamType



from _typeshed import Incomplete
from ubelt.util_mixins import NiceRepr


def split_archive(fpath, ext: str = ...):
    ...


class zopen(NiceRepr):
    fpath: Incomplete
    ext: Incomplete
    name: Incomplete
    mode: Incomplete

    def __init__(self,
                 fpath,
                 mode: str = ...,
                 seekable: bool = ...,
                 ext: str = ...) -> None:
        ...

    @property
    def zfile(self):
        ...

    def namelist(self):
        ...

    def __nice__(self):
        ...

    def __getattr__(self, key):
        ...

    def __dir__(self):
        ...

    def __del__(self) -> None:
        ...

    def __enter__(self):
        ...

    def __exit__(self, *args) -> None:
        ...



r"""
This module exposes the :func:`ubelt.cmd` command, which provides a simple
means for interacting with the commandline.  This uses
:class:`subprocess.Popen` under the hood, but improves upon existing
:mod:`subprocess` functionality by:

(1) Adding the option to "tee" the output, i.e. simultaniously capture and
write to stdout and stderr.

(2) Always specify the command as a string. The :mod:`subprocess` module
expects the command as either a  ``List[str]`` if ``shell=False`` and ``str``
if ``shell=True``. If necessary, :func:`ubelt.util_cmd.cmd` will automatically
convert from one format to the other, so passing in either case will work.

(3) Specificy if the process blocks or not by setting ``detach``. Note: when
``detach is True`` it is not possible to tee the output.

Example:
    >>> import ubelt as ub
    >>> # Running with verbose=1 will write to stdout in real time
    >>> info = ub.cmd('echo "write your command naturally"', verbose=1)
    write your command naturally
    >>> # Unless `detach=True`, `cmd` always returns an info dict.
    >>> print('info = ' + ub.repr2(info))
    info = {
        'command': 'echo "write your command naturally"',
        'cwd': None,
        'err': '',
        'out': 'write your command naturally\n',
        'proc': <...Popen...>,
        'ret': 0,
    }
"""
import sys
import os


__pitch__ = """
The ubelt.cmd command is probably the easiest way to execute a command line program from Python. Unlike os.system, subprocess.check_output, and subprocess.call, The syntax for what you want to call is exactly the same no matter what type of configuration you are using.

Either pass the text you would execute on the command line directly or break it up into a list where each item should be considered its own argument. This works regardless of if shell=True or shell=False, so if your command doesn't work with the safer shell=False, you can turn on shell=True without modifying anything else. You can capture output, print it to the screen, or namely --- something few other packages support --- both (via tee=True or verbose>1).

You can also invoke the call via os.system instead of Popen by setting system=True (although this does come with all of the os.system benefits and restrictions).

I'm biased because I wrote it, but subprocess-tee is the only other package I know of that comes close to getting this right. Maybe invoke?
"""

# import logging
# logging.basicConfig(
#     format='[%(asctime)s %(threadName)s %(levelname)s] %(message)s',
#     level=logging.DEBUG,
#     force=True
# )
# logger = logging.getLogger(__name__)

__all__ = ['cmd']

POSIX = 'posix' in sys.builtin_module_names
WIN32 = sys.platform == 'win32'


def cmd(command, shell=False, detach=False, verbose=0, tee=None, cwd=None,
        env=None, tee_backend='auto', check=False, system=False, timeout=None):
    """
    Executes a command in a subprocess.

    The advantage of this wrapper around subprocess is that
    (1) you control if the subprocess prints to stdout,
    (2) the text written to stdout and stderr is returned for parsing,
    (3) cross platform behavior that lets you specify the command as a string
    or tuple regardless of whether or not shell=True.
    (4) ability to detach, return the process object and allow the process to
    run in the background (eventually we may return a Future object instead).

    Args:
        command (str | List[str]):
            command string, tuple of executable and args, or shell command.

        shell (bool, default=False):
            if True, process is run in shell.

        detach (bool, default=False):
            if True, process is detached and run in background.

        verbose (int, default=0):
            verbosity mode. Can be 0, 1, 2, or 3.

        tee (bool | None):
            if True, simultaneously writes to stdout while capturing output
            from the command. If not specified, defaults to True if verbose >
            0.  If detach is True, then this argument is ignored.

        cwd (str | PathLike | None):
            Path to run command. Defaults to current working directory if
            unspecified.

        env (Dict[str, str] | None):
            environment passed to Popen

        tee_backend (str, default='auto'): backend for tee output.
            Valid choices are: "auto", "select" (POSIX only), and "thread".

        check (bool, default=False):
            if True, check that the return code was zero before returning,
            otherwise raise a :class:`subprocess.CalledProcessError`.
            Does nothing if detach is True.

        system (bool, default=False):
            if True, most other considerations are dropped, and
            :func:`os.system` is used to execute the command in a platform
            dependant way. Other arguments such as env, tee, timeout, and shell
            are all ignored.  (new in version 1.1.0)

        timeout (float):
            If the process does not complete in `timeout` seconds, raises a
            :class:`subprocess.TimeoutExpired`. (new in version 1.1.0)
            Currently unhandled when tee is True.

        log (Callable | None):
            If specified, verbose output is written using this function,
            otherwise the builtin print function is used.

    Returns:
        dict:
            info - information about command status.
            if detach is False ``info`` contains captured standard out,
            standard error, and the return code
            if detach is True ``info`` contains a reference to the process.

    Raises:
        ValueError - on an invalid configuration
        subprocess.TimeoutExpired - if the timeout limit is exceeded
        subprocess.CalledProcessError - if check and the return value is non zero

    Note:
        Inputs can either be text or tuple based. On UNIX we ensure conversion
        to text if shell=True, and to tuple if shell=False. On windows, the
        input is always text based.  See [SO_33560364]_ for a potential
        cross-platform shlex solution for windows.

        When using the tee output, the stdout and stderr may be shuffled from
        what they would be on the command line.

    Related Work:
        https://github.com/pycontribs/subprocess-tee
        https://github.com/mortoray/shelljob
        https://github.com/netinvent/command_runner
        https://www.pyinvoke.org/prior-art.html

    References:
        .. [SO_11495783] https://stackoverflow.com/questions/11495783/redirect-subprocess-stderr-to-stdout
        .. [SO_7729336] https://stackoverflow.com/questions/7729336/how-can-i-print-and-display-subprocess-stdout-and-stderr-output-without-distorti
        .. [SO_33560364] https://stackoverflow.com/questions/33560364/python-windows-parsing-command-lines-with-shlex

    CommandLine:
        xdoctest -m ubelt.util_cmd cmd:6
        python -c "import ubelt as ub; ub.cmd('ping localhost -c 2', verbose=2)"
        pytest "$(python -c 'import ubelt; print(ubelt.util_cmd.__file__)')" -sv --xdoctest-verbose 2

    Example:
        >>> import ubelt as ub
        >>> info = ub.cmd(('echo', 'simple cmdline interface'), verbose=1)
        simple cmdline interface
        >>> assert info['ret'] == 0
        >>> assert info['out'].strip() == 'simple cmdline interface'
        >>> assert info['err'].strip() == ''

    Example:
        >>> import ubelt as ub
        >>> info = ub.cmd('echo str noshell', verbose=0)
        >>> assert info['out'].strip() == 'str noshell'

    Example:
        >>> # windows echo will output extra single quotes
        >>> import ubelt as ub
        >>> info = ub.cmd(('echo', 'tuple noshell'), verbose=0)
        >>> assert info['out'].strip().strip("'") == 'tuple noshell'

    Example:
        >>> # Note this command is formatted to work on win32 and unix
        >>> import ubelt as ub
        >>> info = ub.cmd('echo str&&echo shell', verbose=0, shell=True)
        >>> assert info['out'].strip() == 'str' + chr(10) + 'shell'

    Example:
        >>> import ubelt as ub
        >>> info = ub.cmd(('echo', 'tuple shell'), verbose=0, shell=True)
        >>> assert info['out'].strip().strip("'") == 'tuple shell'

    Example:
        >>> import pytest
        >>> import ubelt as ub
        >>> info = ub.cmd('echo hi', check=True)
        >>> import subprocess
        >>> with pytest.raises(subprocess.CalledProcessError):
        >>>     ub.cmd('exit 1', check=True, shell=True)

    Example:
        >>> import ubelt as ub
        >>> from os.path import join, exists
        >>> dpath = ub.Path.appdir('ubelt', 'test').ensuredir()
        >>> fpath1 = (dpath / 'cmdout1.txt').delete()
        >>> fpath2 = (dpath / 'cmdout2.txt').delete()
        >>> # Start up two processes that run simultaneously in the background
        >>> info1 = ub.cmd(('touch', str(fpath1)), detach=True)
        >>> info2 = ub.cmd('echo writing2 > ' + str(fpath2), shell=True, detach=True)
        >>> # Detached processes are running in the background
        >>> # We can run other code while we wait for them.
        >>> while not exists(fpath1):
        ...     pass
        >>> while not exists(fpath2):
        ...     pass
        >>> # communicate with the process before you finish
        >>> # (otherwise you may leak a text wrapper)
        >>> info1['proc'].communicate()
        >>> info2['proc'].communicate()
        >>> # Check that the process actually did finish
        >>> assert (info1['proc'].wait()) == 0
        >>> assert (info2['proc'].wait()) == 0
        >>> # Check that the process did what we expect
        >>> assert fpath1.read_text() == ''
        >>> assert fpath2.read_text().strip() == 'writing2'

    Example:
        >>> # Can also use ub.cmd to call os.system
        >>> import pytest
        >>> import ubelt as ub
        >>> import subprocess
        >>> info = ub.cmd('echo hi', check=True, system=True)
        >>> with pytest.raises(subprocess.CalledProcessError):
        >>>     ub.cmd('exit 1', check=True, shell=True)
    """
    # In the future we might allow the user to pass a custom log function
    # But this has weird interactions with how the tee process works
    # becasue of the assumption stdout.write does not emit a newline
    log = print

    import subprocess
    # TODO: stdout, stderr - experimental - custom file to pipe stdout/stderr to
    # Determine if command is specified as text or a tuple
    if isinstance(command, str):
        command_text = command
        command_tup = None
    else:
        import pipes
        command_parts = []
        # Allow the user to specify paths as part of the command
        for part in command:
            if isinstance(part, os.PathLike):
                part = os.fspath(part)
            command_parts.append(part)
        command_tup = tuple(command_parts)
        command_text = ' '.join(list(map(pipes.quote, command_tup)))

    if shell or sys.platform.startswith('win32'):
        # When shell=True, args is sent to the shell (e.g. bin/sh) as text
        args = command_text
    else:
        # When shell=False, args is a list of executable and arguments
        if command_tup is None:
            # parse this out of the string
            # NOTE: perhaps use the solution from [SO_33560364] here?
            import shlex
            command_tup = shlex.split(command_text)
            # command_tup = shlex.split(command_text, posix=not WIN32)
        args = command_tup

    if tee is None:
        tee = verbose > 0

    if tee and tee_backend not in {'auto', 'thread', 'select'}:
        raise ValueError('tee_backend must be select, thread, or auto')

    if verbose > 1:
        import platform
        import getpass
        from ubelt import shrinkuser
        if verbose > 2:
            try:
                log('┌─── START CMD ───')
            except Exception:  # nocover
                log('+=== START CMD ===')
        cwd_ = os.getcwd() if cwd is None else cwd
        compname = platform.node()
        username = getpass.getuser()
        cwd_ = shrinkuser(cwd_)
        ps1 = '[ubelt.cmd] {}@{}:{}$ '.format(username, compname, cwd_)
        log(ps1 + command_text)

    # Create a new process to execute the command
    def make_proc():
        # delay the creation of the process until we validate all args
        proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, shell=shell,
                                universal_newlines=True, cwd=cwd, env=env)
        return proc

    if system:
        info = {
            'command': command_text,
            'out': None,
            'err': None,
        }
        ret = os.system(command_text)
        info['ret'] = ret
    elif detach:
        info = {'proc': make_proc(), 'command': command_text}
        if verbose > 0:  # nocover
            log('...detaching')
    else:
        if tee:
            # We logging stdout and stderr, while simulaniously piping it to
            # another stream.
            stdout = sys.stdout
            stderr = sys.stderr
            proc = make_proc()
            with proc:
                out, err = _tee_output(
                    proc=proc, stdout=stdout, stderr=stderr,
                    backend=tee_backend, timeout=timeout,
                    command_text=command_text)
                (out_, err_) = proc.communicate(timeout=timeout)
        else:
            proc = make_proc()
            with proc:
                try:
                    (out, err) = proc.communicate(timeout=timeout)
                except subprocess.TimeoutExpired as exc:
                    # Follow the error handling in the stdlib implementaton of
                    # subprocess.run
                    proc.kill()
                    if WIN32:  # nocover
                        # Win32 needs a communicate after the kill to get the
                        # output. See stdlib for details.
                        exc.stdout, exc.stderr = proc.communicate()
                    else:
                        # Posix implementations already handle the populate.
                        proc.wait()
                    raise
        # We used the popen context manager, which means that wait was called,
        # the process has existed, so it is safe to return a reference to the
        # process object.
        ret = proc.poll()
        info = {
            'out': out,
            'err': err,
            'ret': ret,
            'proc': proc,
            'cwd': cwd,
            'command': command_text
        }

    if not detach:
        if verbose > 2:
            # https://en.wikipedia.org/wiki/Box-drawing_character
            try:
                log('└─── END CMD ───')
            except Exception:  # nocover
                log('L___ END CMD ___')

        if check:
            if info['ret'] != 0:
                raise subprocess.CalledProcessError(
                    info['ret'], info['command'], info['out'], info['err'])
    return info


def _textio_iterlines(stream):
    """
    Iterates over lines in a TextIO stream until an EOF is encountered.
    This is the iterator version of stream.readlines()

    Args:
        stream (io.TextIOWrapper): The stream to finish reading.

    Yields:
        str: a line read from the stream.
    """
    try:
        # These if statements help mitigate race conditions but does not solve
        # them if the stream closes in the middle of a readline.
        if stream.closed:  # nocover
            return
        line = stream.readline()
        while line != '':
            yield line
            if stream.closed:  # nocover
                return
            line = stream.readline()
    except ValueError:  # nocover
        # Ignore I/O operation on closed files, the process was likely
        # killed.
        raise
        ...


def _proc_async_iter_stream(proc, stream, buffersize=1, timeout=None):
    """
    Reads output from a process in a separate thread

    Args:
        proc (subprocess.Popen): The process being run

        stream (io.TextIOWrapper): A stream belonging to the process
            e.g. ``proc.stdout`` or ``proc.stderr``.

        buffersize (int): Size of the returned queue.

    Returns:
        queue.Queue:
            The queue that the output lines will be asynchronously written to
            as they are read from the stream.
    """
    import queue
    import threading
    # logger.debug(f"Create and start thread for {id(stream)}")
    out_queue = queue.Queue(maxsize=buffersize)
    control_queue = queue.Queue(maxsize=1)
    io_thread = threading.Thread(
        target=_enqueue_output_thread_worker, args=(
            proc, stream, out_queue, control_queue, timeout))
    io_thread.daemon = True  # thread dies with the program
    io_thread.start()
    return io_thread, out_queue, control_queue


def _enqueue_output_thread_worker(proc, stream, out_queue, control_queue, timeout=None):
    """
    Thread worker function

    This follows a similar strategy employed in
    http://eyalarubas.com/python-subproc-nonblock.html and
    https://stackoverflow.com/questions/375427/a-non-blocking-read-on-a-subprocess-pipe-in-python/4896288#4896288

    Args:
        proc (subprocess.Popen): The process being run

        stream (io.TextIOWrapper): A stream belonging to the process
            e.g. ``proc.stdout`` or ``proc.stderr``.

        out_queue (queue.Queue): The queue to write to.

        control_queue (queue.Queue): For sending a signal to stop the thread

        timeout (None | float): amount of time to allow before stopping
    """
    import queue
    # logger.debug(f"Start worker for {id(stream)=} with {timeout=}")

    def _check_if_stopped():  # nocover
        try:
            # Check if we were told to stop
            control_queue.get_nowait()
        except queue.Empty:
            ...
        else:
            # logger.debug(f"Thread acknowledges stop request for {id(stream)}")
            return True

    def enqueue(item):  # nocover
        # Alternate between checking if we were stopped and putting the item in
        # the queue. This helps with the issue of an open process stream on
        # exit but it doesn't fully solve the issue because we still might
        # block on the stream.readline, therefore we can't guarentee this
        # thread will exit before the process does.
        if timeout is None:
            # If timeout is None, we can optimize this and just use the
            # blocking call.
            out_queue.put(item)
            return True

        # logger.debug('Waiting to put in item')
        while True:
            if _check_if_stopped():
                return False
            try:
                out_queue.put(item, block=False)
                # logger.debug('Thread put in item')
            except queue.Full:
                pass
            else:
                return True

    while proc.poll() is None:

        # Note: if the underlying process has buffered output, we may get this
        # line well after it is initially emitted and thus be stuck waiting
        # here for some time.
        # logger.debug(f"ENQUEUE Waiting for line {id(stream)}")
        line = stream.readline()

        # logger.debug(f"ENQUEUE LIVE {id(stream)} {line!r}")
        if not enqueue(line):  # nocover
            return

    if _check_if_stopped():  # nocover
        return

    # Coverage note: on Python 3.10 it seems like the tests dont always cover
    # these lines. We don't have much control over if this happens or not, so
    # we will exclude them from coverage checks.
    for line in _textio_iterlines(stream):  # nocover
        # logger.debug(f"ENQUEUE FINAL {id(stream)} {line!r}")
        if not enqueue(line):  # nocover
            return

    # logger.debug(f"STREAM IS DONE {id(stream)}")
    # signal that the stream is finished
    if not enqueue(None):  # nocover
        return


def _proc_iteroutput_thread(proc, timeout=None):
    """
    Iterates over output from a process line by line.

    Follows the answers from [SO_375427]_.

    Note:
        WARNING. Current implementation might have bugs with other threads.
        This behavior was seen when using earlier versions of tqdm. I'm not
        sure if this was our bug or tqdm's. Newer versions of tqdm fix this,
        but I cannot guarantee that there isn't an issue on our end.

    Yields:
        Tuple[str, str]: oline, eline - stdout and stderr line

    References:
        .. [SO_375427] https://stackoverflow.com/questions/375427/non-blocking-read-subproc
    """
    import queue

    # logger.debug("Create stdout/stderr streams")
    # Create threads that read stdout / stderr and queue up the output
    stdout_thread, stdout_queue, stdout_ctrl = _proc_async_iter_stream(proc, proc.stdout, timeout=timeout)
    stderr_thread, stderr_queue, stderr_ctrl = _proc_async_iter_stream(proc, proc.stderr, timeout=timeout)

    stdout_live = True
    stderr_live = True

    if timeout is not None:
        from time import monotonic as _time
        import subprocess
        start_time = _time()

    # read from the output asynchronously until
    while stdout_live or stderr_live:
        # Note: This function loop happens very quickly.
        # # logger.debug("Fast loop: check stdout / stderr threads")

        if timeout is not None:
            # Check for timeouts
            elapsed = _time() - start_time
            if elapsed >= timeout:
                stdout_ctrl.put('STOP')
                stderr_ctrl.put('STOP')
                # Unfortunately we can't guarentee that the threads will stop
                # because they might get stuck in a readline
                # stdout_thread.join()
                # stderr_thread.join()
                yield subprocess.TimeoutExpired, subprocess.TimeoutExpired

        if stdout_live:  # pragma: nobranch
            try:
                oline = stdout_queue.get_nowait()
                stdout_live = oline is not None
            except queue.Empty:
                oline = None
        if stderr_live:
            try:
                eline = stderr_queue.get_nowait()
                stderr_live = eline is not None
            except queue.Empty:
                eline = None
        if oline is not None or eline is not None:
            yield oline, eline


def _proc_iteroutput_select(proc, timeout=None):
    """
    Iterates over output from a process line by line

    UNIX only. Use :func:`_proc_iteroutput_thread` instead for a cross platform
    solution based on threads.

    Args:
        proc (subprocess.Popen): the process being run

        timeout (None | float): amount of time to allow before stopping

    Yields:
        Tuple[str, str]: oline, eline - stdout and stderr line
    """
    from itertools import zip_longest
    import select

    if timeout is not None:
        from time import monotonic as _time
        import subprocess
        start_time = _time()

    # Read output while the external program is running
    while proc.poll() is None:
        if timeout is not None:
            elapsed = _time() - start_time
            if elapsed >= timeout:
                yield subprocess.TimeoutExpired, subprocess.TimeoutExpired
                return  # nocover

        reads = [proc.stdout.fileno(), proc.stderr.fileno()]
        ret = select.select(reads, [], [], timeout)
        oline = eline = None
        for fd in ret[0]:
            if fd == proc.stdout.fileno():
                oline = proc.stdout.readline()
            if fd == proc.stderr.fileno():
                eline = proc.stderr.readline()
        yield oline, eline

    # Grab any remaining data in stdout and stderr after the process finishes
    oline_iter = _textio_iterlines(proc.stdout)
    eline_iter = _textio_iterlines(proc.stderr)
    for oline, eline in zip_longest(oline_iter, eline_iter):
        yield oline, eline


def _tee_output(proc, stdout=None, stderr=None, backend='thread',
                timeout=None, command_text=None):
    """
    Simultaneously reports and captures stdout and stderr from a process

    subprocess must be created using (stdout=subprocess.PIPE,
    stderr=subprocess.PIPE)

    Args:
        proc (subprocess.Popen): the process being run

        stdout (io.TextIOWrapper): typically sys.stdout

        stderr (io.TextIOWrapper): typically sys.stderr

        backend (str): thread, select or auto

        timeout (None | float): time before raising a timeout error

        command_text (str): used only to construct a TimeoutExpired error.

    Returns:
        Tuple[str, str]: recorded stdout and stderr
    """
    import subprocess
    logged_out = []
    logged_err = []
    if backend == 'auto':
        # backend = 'select' if POSIX else 'thread'
        backend = 'thread'

    if backend == 'select':
        if not POSIX:  # nocover
            raise NotImplementedError('select is only available on posix')
        # the select-based version is stable, but slow
        _proc_iteroutput = _proc_iteroutput_select
    elif backend == 'thread':
        # the thread version is fast, but might run into issues.
        _proc_iteroutput = _proc_iteroutput_thread
    else:  # nocover
        # The value of "backend" should be checked before we create the
        # processes, otherwise we will have a dangling process
        raise AssertionError(
            'Invalid backend, but the check should have already a happened')

    # TODO: handle timeout
    output_gen = _proc_iteroutput(proc, timeout=timeout)
    # logger.debug("Start waiting for buffered output")
    for oline, eline in output_gen:
        if timeout is not None:
            if oline is subprocess.TimeoutExpired or eline is subprocess.TimeoutExpired:
                # logger.error("Timeout error triggered!")
                try:
                    out = ''.join(logged_out)
                except UnicodeDecodeError:  # nocover
                    out = '\n'.join(_.decode('utf-8') for _ in logged_out)
                try:
                    err = ''.join(logged_err)
                except UnicodeDecodeError:  # nocover
                    err = '\n'.join(_.decode('utf-8') for _ in logged_err)
                # Following the standard library implementation of
                # :func:`subprocess.run`, we kill (not terminate) the process
                # when the timeout expires. We shouldn't need the extra
                # communicate fix for windows because we report the tee-ed
                # output that already exists. But lets see what the CI says.
                proc.kill()
                proc.wait()
                raise subprocess.TimeoutExpired(command_text, timeout, out, err)
        if oline:
            # logger.debug("Write oline to stdout.write and logged_out")
            if stdout:  # pragma: nobranch
                stdout.write(oline)
                stdout.flush()
            logged_out.append(oline)
        if eline:
            # logger.debug("Write eline to stderr.write and logged_err")
            if stderr:  # pragma: nobranch
                stderr.write(eline)
                stderr.flush()
            logged_err.append(eline)
        # logger.debug("Continue waiting for buffered output")

    # The motivation for this logic is unclear.
    # In what cases is the logged output returned as bytes or text?
    # Using a bytes join probably makes more sense in most cases.
    try:
        out = ''.join(logged_out)
    except UnicodeDecodeError:  # nocover
        out = '\n'.join(_.decode('utf-8') for _ in logged_out)
    try:
        err = ''.join(logged_err)
    except UnicodeDecodeError:  # nocover
        err = '\n'.join(_.decode('utf-8') for _ in logged_err)

    return out, err


# Stub for possible object oriented interface
# class Command:
#     """
#     TODO
#     """
#     ...
